import torch


def update_base(layer, tensor_name):
    with torch.no_grad():
        setattr(layer, tensor_name, getattr(layer, tensor_name).data)


def transpose(fun):
    def new_fun(self, X, *args, **kwargs):
        # It might happen that we get at tuple inside ``right_inverse``
        # In that case we do nothing
        if isinstance(X, torch.Tensor):
            if self.transposed:
                X = X.transpose(-2, -1)
        X = fun(self, X, *args, **kwargs)
        if self.transposed:
            X = X.transpose(-2, -1)
        return X

    return new_fun


def _extra_repr(**kwargs):  # noqa: C901
    if "n" in kwargs:
        ret = "n={}".format(kwargs["n"])
    elif "dim" in kwargs:
        ret = "dim={}".format(kwargs["dim"])
    else:
        ret = ""

    if "k" in kwargs:
        ret += ", k={}".format(kwargs["k"])
    if "rank" in kwargs:
        ret += ", rank={}".format(kwargs["rank"])
    if "radius" in kwargs:
        ret += ", radius={}".format(kwargs["radius"])
    if "lam" in kwargs:
        ret += ", lambda={}".format(kwargs["lam"])
    if "f" in kwargs:
        ret += ", f={}".format(kwargs["f"].__name__)
    if "tensorial_size" in kwargs:
        ts = kwargs["tensorial_size"]
        if len(ts) != 0:
            ret += ", tensorial_size={}".format(tuple(ts))
    if "triv" in kwargs:
        ret += ", triv={}".format(kwargs["triv"].__name__)
    if "no_inv" in kwargs:
        if kwargs["no_inv"]:
            ret += ", no inverse"
    if "transposed" in kwargs:
        if kwargs["transposed"]:
            ret += ", transposed"
    return ret



from .constraints import (
    sphere,
    skew,
    symmetric,
    orthogonal,
    grassmannian,
    almost_orthogonal,
    low_rank,
    fixed_rank,
    invertible,
    positive_definite,
    positive_semidefinite,
    positive_semidefinite_low_rank,
    positive_semidefinite_fixed_rank,
)
from .product import ProductManifold
from .reals import Rn
from .skew import Skew
from .symmetric import Symmetric
from .so import SO
from .sphere import Sphere, SphereEmbedded
from .stiefel import Stiefel
from .grassmannian import Grassmannian
from .almostorthogonal import AlmostOrthogonal
from .lowrank import LowRank
from .fixedrank import FixedRank
from .glp import GLp
from .psd import PSD
from .pssd import PSSD
from .pssdfixedrank import PSSDFixedRank
from .pssdlowrank import PSSDLowRank
from .utils import update_base


__version__ = "0.2.0"


__all__ = [
    "ProductManifold",
    "Grassmannian",
    "LowRank",
    "Rn",
    "Skew",
    "Symmetric",
    "SO",
    "Sphere",
    "SphereEmbedded",
    "Stiefel",
    "AlmostOrthogonal",
    "GLp",
    "FixedRank",
    "PSD",
    "PSSD",
    "PSSDLowRank",
    "PSSDFixedRank",
    "skew",
    "symmetric",
    "sphere",
    "orthogonal",
    "grassmannian",
    "low_rank",
    "fixed_rank",
    "almost_orthogonal",
    "invertible",
    "positive_definite",
    "positive_semidefinite",
    "positive_semidefinite_low_rank",
    "positive_semidefinite_fixed_rank",
    "update_base",
]



import torch
from torch import nn
from functools import partial

try:
    from torch.linalg import eigh
    from torch.linalg import eigvalsh
except ImportError:
    from torch import symeig

    eigh = partial(symeig, eigenvectors=True)

    def eigvalsh(X):
        return symeig(X, eigenvectors=False).eigenvalues


from .product import ProductManifold
from .stiefel import Stiefel
from .reals import Rn
from .exceptions import (
    VectorError,
    NonSquareError,
    RankError,
    InManifoldError,
    InverseError,
)
from .utils import _extra_repr


def _decreasing_eigh(X, eigenvectors):
    if eigenvectors:
        L, Q = eigh(X)
        return L.flip(-1), Q.flip(-1)
    else:
        return eigvalsh(X).flip(-1)


class Symmetric(nn.Module):
    def __init__(self, lower=True):
        r"""
        Vector space of symmetric matrices, parametrized in terms of the upper or lower
        triangular part of a matrix.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            lower (bool): Optional. Uses the lower triangular part of the matrix to
                parametrize the matrix. Default: ``True``
        """
        super().__init__()
        self.lower = lower

    @staticmethod
    def frame(X, lower):
        if lower:
            return X.tril(0) + X.tril(-1).transpose(-2, -1)
        else:
            return X.triu(0) + X.triu(1).transpose(-2, -1)

    def forward(self, X):
        if len(X.size()) < 2:
            raise VectorError(type(self).__name__, X.size())
        if X.size(-2) != X.size(-1):
            raise NonSquareError(type(self).__name__, X.size())
        return self.frame(X, self.lower)

    @staticmethod
    def in_manifold(X, eps=1e-6):
        return (
            X.dim() >= 2
            and X.size(-2) == X.size(-1)
            and torch.allclose(X, X.transpose(-2, -1), atol=eps)
        )


class SymF(ProductManifold):
    def __init__(self, size, rank, f, triv="expm"):
        r"""
        Space of the symmetric matrices of rank at most k with eigenvalues
        in the image of a given function

        Args:
            size (torch.size): Size of the tensor to be parametrized
            rank (int): Rank of the matrices.
                It has to be less or equal to
                :math:`\min(\texttt{size}[-1], \texttt{size}[-2])`
            f (callable or pair of callables): Either:

                - A callable

                - A pair of callables such that the second is a (right)
                  inverse of the first
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`Q` in the eigenvalue
                decomposition. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        n, tensorial_size = SymF.parse_size(size)
        if rank > n or rank < 1:
            raise RankError(n, n, rank)
        super().__init__(SymF.manifolds(n, rank, tensorial_size, triv))
        self.n = n
        self.tensorial_size = tensorial_size
        self.rank = rank
        f, inv = SymF.parse_f(f)
        self.f = f
        self.inv = inv

    @classmethod
    def parse_size(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        n, k = size[-2:]
        tensorial_size = size[:-2]
        if n != k:
            raise NonSquareError(cls.__name__, size)
        return n, tensorial_size

    @staticmethod
    def parse_f(f):
        if callable(f):
            return f, None
        elif isinstance(f, tuple) and callable(f[0]) and callable(f[1]):
            return f
        else:
            raise ValueError(
                "Argument f is not callable nor a pair of callables. "
                "Found {}".format(f)
            )

    @staticmethod
    def manifolds(n, rank, tensorial_size, triv):
        size_q = tensorial_size + (n, rank)
        size_l = tensorial_size + (rank,)
        return Stiefel(size_q, triv=triv), Rn(size_l)

    def frame(self, X):
        L = X.diagonal(dim1=-2, dim2=-1)[..., : self.rank]
        X = X[..., : self.rank]
        return X, L

    def submersion(self, Q, L):
        L = self.f(L)
        Qt = Q.transpose(-2, -1)
        # Multiply the three of them as Q\LambdaQ^T
        return Q @ (L.unsqueeze(-1).expand_as(Qt) * Qt)

    def forward(self, X):
        X = self.frame(X)
        Q, L = super().forward(X)
        return self.submersion(Q, L)

    def frame_inv(self, X1, X2):
        size = self.tensorial_size + (self.n, self.n)
        ret = torch.zeros(*size, dtype=X1.dtype, device=X1.device)
        with torch.no_grad():
            ret[..., : self.rank] += X1
            ret[..., : self.rank, : self.rank] += torch.diag_embed(X2)
        return ret

    def submersion_inv(self, X, check_in_manifold=True):
        if isinstance(X, torch.Tensor):
            with torch.no_grad():
                L, Q = _decreasing_eigh(X, eigenvectors=True)
            if check_in_manifold and not self.in_manifold_eigen(L):
                raise InManifoldError(X, self)
        else:
            # We assume that we got the L, Q factorized in a tuple / list
            L, Q = X
            if check_in_manifold and not self.in_manifold_tuple(L, Q):
                raise InManifoldError(X, self)
        if self.inv is None:
            raise InverseError(self)
        with torch.no_grad():
            Q = Q[..., : self.rank]
            L = L[..., : self.rank]
            # Multiply the three of them as Q\LambdaQ^T
            L = self.inv(L)
        return L, Q

    def right_inverse(self, X, check_in_manifold=True):
        L, Q = self.submersion_inv(X, check_in_manifold)
        X1, X2 = super().right_inverse([Q, L], check_in_manifold=False)
        return self.frame_inv(X1, X2)

    def in_manifold_eigen(self, L, eps=1e-6):
        r"""
        Checks that an ordered vector of eigenvalues values is in the manifold.

        For tensors with more than 1 dimension the first dimensions are
        treated as batch dimensions.

        Args:
            L (torch.Tensor): Vector of eigenvalues
            eps (float): Optional. Threshold at which the eigenvalues are
                considered to be zero
                Default: ``1e-6``
        """
        if L.size()[:-1] != self.tensorial_size:
            return False
        if L.size(-1) > self.rank:
            # We compute the \infty-norm of the remaining dimension
            D = L[..., self.rank :]
            infty_norm_err = D.abs().max(dim=-1).values
            if (infty_norm_err > 5.0 * eps).any():
                return False
        return (L[..., : self.rank] >= -eps).all().item()

    def in_manifold_tuple(self, L, Q, eps=1e-6):
        return (
            self.in_manifold_eigen(L, eps)
            and self[0].in_manifold(Q)
            and self[1].in_manifold(L)
        )

    def in_manifold(self, X, eps=1e-6):
        r"""
        Checks that a matrix is in the manifold. The matrix may be given factorized
        as a pair :math:`(\Lambda, Q)` with :math:`\Lambda` a vector of eigenvalues
        and :math:`Q` a matrix of eigenvectors.

        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor or tuple): The matrix to be checked or a tuple
                ``(eigenvectors, eigenvalues)`` as returned by ``torch.linalg.eigh``
                or ``self.sample(factorized=True)``.
            eps (float): Optional. Threshold at which the singular values are
                    considered to be zero
                    Default: ``1e-6``
        """
        if isinstance(X, tuple):
            if len(X) == 2:
                return self.in_manifold_tuple(X[0], X[1])
            else:
                return False
        size = self.tensorial_size + (self.n, self.n)
        if X.size() != size or not Symmetric.in_manifold(X, eps):
            return False

        L = _decreasing_eigh(X, eigenvectors=False)
        return self.in_manifold_eigen(L, eps)

    def sample(self, init_=torch.nn.init.xavier_normal_, factorized=True):
        r"""
        Returns a randomly sampled matrix on the manifold as

        .. math::

            WW^\intercal \qquad W_{i,j} \sim \texttt{init_}

        By default ``init\_`` is a (xavier) normal distribution, so that the
        returned matrix follows a Wishart distribution.

        The output of this method can be used to initialize a parametrized tensor
        that has been parametrized with this or any other manifold as::

            >>> layer = nn.Linear(20, 20)
            >>> M = PSSD(layer.weight.size())
            >>> geotorch.register_parametrization(layer, "weight", M)
            >>> layer.weight = M.sample()

        Args:
            init\_ (callable): Optional.
                    A function that takes a tensor and fills it in place according
                    to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: ``torch.nn.init.xavier_normal_``
            factorized (bool): Optional. Return an eigenvalue decomposition of the
                    sampled matrix as a tuple :math:`(\Lambda, Q)`.
                    Using ``factorized=True`` is more efficient when the result is
                    used to initialize a parametrized tensor.
                    Default: ``True``
        """
        with torch.no_grad():
            device = self[0].base.device
            dtype = self[0].base.dtype
            X = torch.empty(
                *(self.tensorial_size + (self.n, self.n)), device=device, dtype=dtype
            )
            init_(X)
            X = X @ X.transpose(-2, -1)
            L, Q = _decreasing_eigh(X, eigenvectors=True)
            L = L[..., : self.rank]
            Q = Q[..., : self.rank]
            if factorized:
                return L, Q
            else:
                # Project onto the manifold
                Qt = Q.transpose(-2, -1)
                # Multiply the three of them as Q\LambdaQ^T
                return Q @ (L.unsqueeze(-1).expand_as(Qt) * Qt)

    def extra_repr(self):
        return _extra_repr(
            n=self.n,
            rank=self.rank,
            tensorial_size=self.tensorial_size,
            f=self.f,
            no_inv=self.inv is None,
        )



from unittest import TestCase

from geotorch.pssdlowrank import PSSDLowRank
from geotorch.pssdfixedrank import PSSDFixedRank
from geotorch.pssd import PSSD
from geotorch.psd import PSD


class TestPSSDLowRank(TestCase):
    def test_positive_semidefinite_errors(self):
        for cls in [PSSDLowRank, PSSDFixedRank]:
            # rank always has to be 1 <= rank <= n
            with self.assertRaises(ValueError):
                cls(size=(4, 4), rank=5)
            with self.assertRaises(ValueError):
                cls(size=(3, 3), rank=0)
            # Instantiate it in a non-square matrix
            with self.assertRaises(ValueError):
                cls(size=(3, 6), rank=2)
            # Try to instantiate it in a vector rather than a matrix
            with self.assertRaises(ValueError):
                cls(size=(5,), rank=1)

        for cls in [PSSD, PSD]:
            # Try to instantiate it in a vector rather than a matrix
            with self.assertRaises(ValueError):
                cls(size=(5,))
            # Or a non-square
            with self.assertRaises(ValueError):
                cls(size=(5, 3))

        # Pass a non-callable object
        with self.assertRaises(ValueError):
            PSSDFixedRank(size=(5, 2), rank=1, f=3)
        # Or the wrong string
        with self.assertRaises(ValueError):
            PSSDFixedRank(size=(5, 3), rank=2, f="fail")
        # Same with PSD
        with self.assertRaises(ValueError):
            PSD(size=(5, 2), f=3)
        with self.assertRaises(ValueError):
            PSD(size=(5, 3), f="fail")



from unittest import TestCase

from geotorch.glp import GLp


class TestGLp(TestCase):
    def test_GLp_errors(self):
        # Non square
        with self.assertRaises(ValueError):
            GLp(size=(4, 3))
        # Try to instantiate it in a vector rather than a matrix
        with self.assertRaises(ValueError):
            GLp(size=(5,))



# Tests for the Sphere
from unittest import TestCase
import itertools

import torch
import torch.nn as nn

import geotorch.parametrize as P
from geotorch.symmetric import Symmetric, SymF


class TestSymmetric(TestCase):
    def test_backprop(self):
        r"""Test that we may instantiate the parametrizations and
        register them in modules of several sizes. Check that the
        results are on the sphere
        """
        sizes = [1, 2, 3, 8]

        for n, lower in itertools.product(sizes, [True, False]):
            layer = nn.Linear(n, n)
            P.register_parametrization(layer, "weight", Symmetric(lower=lower))

            input_ = torch.rand(5, n)
            optim = torch.optim.SGD(layer.parameters(), lr=1.0)

            # Assert that is stays in Sym(n) after some optimiser steps
            for _ in range(2):
                with P.cached():
                    self.assertTrue(Symmetric.in_manifold(layer.weight))
                    loss = layer(input_).sum()
                optim.zero_grad()
                loss.backward()
                optim.step()

    def test_construction(self):
        # Non-square sym
        with self.assertRaises(ValueError):
            Symmetric()(torch.rand(3, 2))

        with self.assertRaises(ValueError):
            Symmetric()(torch.rand(1, 3))

        # Try to instantiate it in a vector rather than a matrix
        with self.assertRaises(ValueError):
            Symmetric()(torch.rand(4))

        # Instantiate it with a non-callable object
        with self.assertRaises(ValueError):
            SymF(size=(4, 4), rank=4, f=3.0)
        # Or with the wrong rank
        with self.assertRaises(ValueError):
            SymF(size=(4, 4), rank=5, f=lambda: None)
        with self.assertRaises(ValueError):
            SymF(size=(4, 4), rank=0, f=lambda: None)
        # Or on vectors
        with self.assertRaises(ValueError):
            SymF(size=(4,), rank=2, f=lambda: None)
        # Or on non-square matrices
        with self.assertRaises(ValueError):
            SymF(size=(4, 3), rank=2, f=lambda: None)

    def test_repr(self):
        print(Symmetric())



import torch
from .lowrank import LowRank
from .exceptions import VectorError, InManifoldError, InverseError
from .utils import _extra_repr


def scaled_sigmoid(t):
    return 2.0 * (torch.sigmoid(t) - 0.5)


def inv_scaled_sigmoid(t):
    y = 0.5 * t + 0.5
    return torch.log(y / (1.0 - y))


class AlmostOrthogonal(LowRank):
    fs = {
        "scaled_sigmoid": (scaled_sigmoid, inv_scaled_sigmoid),
        "tanh": (torch.tanh, torch.atanh),
        "sin": (torch.sin, torch.asin),
    }

    def __init__(self, size, lam, f="sin", triv="expm"):
        r"""Manifold of matrices with singular values in the interval
        :math:`(1-\lambda, 1+\lambda)`.

        The possible default maps are the :math:`\sin,\,\tanh` functions and a scaled
        sigmoid. The sigmoid is scaled as
        :math:`\operatorname{scaled\_sigmoid}(x) = 2\sigma(x) - 1`
        where :math:`\sigma` is the usual sigmoid function.
        This is done so that the image of the scaled sigmoid is :math:`(-1, 1)`.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            lam (float): Radius of the interval. A float in the interval :math:`(0, 1]`
            f (str or callable or pair of callables): Optional. Either:

                - One of ``["scaled_sigmoid", "tanh", "sin"]``

                - A callable that maps real numbers to the interval :math:`(-1, 1)`

                - A pair of callables such that the first maps the real numbers to
                  :math:`(-1, 1)` and the second is a (right) inverse of the first

                Default: ``"sin"``
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`U` and :math:`V` in
                the SVD. It can be one of ``["expm", "cayley"]`` or a custom callable.
                Default: ``"expm"``

        """
        super().__init__(size, AlmostOrthogonal.rank(size), triv=triv)
        if lam < 0.0 or lam > 1.0:
            raise ValueError("The radius has to be between 0 and 1. Got {}".format(lam))
        self.lam = lam
        f, inv = AlmostOrthogonal.parse_f(f)
        self.f = f
        self.inv = inv

    @staticmethod
    def parse_f(f):
        if f in AlmostOrthogonal.fs.keys():
            return AlmostOrthogonal.fs[f]
        elif callable(f):
            return f, None
        elif isinstance(f, tuple) and callable(f[0]) and callable(f[1]):
            return f
        else:
            raise ValueError(
                "Argument f was not recognized and is "
                "not callable or a pair of callables. "
                "Should be one of {}. Found {}".format(
                    list(AlmostOrthogonal.fs.keys()), f
                )
            )

    @classmethod
    def rank(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        return min(*size[-2:])

    def submersion(self, U, S, V):
        S = 1.0 + self.lam * self.f(S)
        return super().submersion(U, S, V)

    def submersion_inv(self, X, check_in_manifold=True):
        if self.inv is None:
            raise InverseError(self)
        U, S, V = super().submersion_inv(X)
        if check_in_manifold and not self.in_manifold_singular_values(S):
            raise InManifoldError(X, self)
        # Harcoded epsilon... not a good practice
        if self.lam < 1e-6:
            S = S - 1.0
        else:
            S = self.inv((S - 1.0) / self.lam)
        return U, S, V

    def in_manifold_singular_values(self, S, eps=1e-5):
        lam = self.lam
        if self.lam <= eps:
            lam = eps
        return (
            super().in_manifold_singular_values(S, eps)
            and ((S - 1.0).abs() <= lam).all().item()
        )

    def sample(self, distribution="uniform", init_=None, factorized=True):
        r"""
        Returns a randomly sampled orthogonal matrix according to the specified
        ``distribution``. The options are:

            - ``"uniform"``: Samples a tensor distributed according to the Haar measure
              on :math:`\operatorname{SO}(n)`

            - ``"torus"``: Samples a block-diagonal skew-symmetric matrix.
              The blocks are of the form
              :math:`\begin{pmatrix} 0 & b \\ -b & 0\end{pmatrix}` where :math:`b` is
              distributed according to ``init_``. This matrix will be then projected
              onto :math:`\operatorname{SO}(n)` using ``self.triv``

        .. note

            The ``"torus"`` initialization is particularly useful in recurrent kernels
            of RNNs

        The output of this method can be used to initialize a parametrized tensor
        that has been parametrized with this or any other manifold as::

            >>> layer = nn.Linear(20, 20)
            >>> M = AlmostOrthogonal(layer.weight.size(), lam=0.5)
            >>> geotorch.register_parametrization(layer, "weight", M)
            >>> layer.weight = M.sample()

        Args:
            distribution (string): Optional. One of ``["uniform", "torus"]``.
                    Default: ``"uniform"``
            init\_ (callable): Optional. To be used with the ``"torus"`` option.
                    A function that takes a tensor and fills it in place according
                    to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: :math:`\operatorname{Uniform}(-\pi, \pi)`
            factorized (bool): Optional. Return an SVD decomposition of the
                    sampled matrix as a tuple :math:`(U, \Sigma, V)`.
                    Using ``factorized=True`` is more efficient when the result is
                    used to initialize a parametrized tensor.
                    Default: ``True``
        """
        with torch.no_grad():
            device = self[0].base.device
            dtype = self[0].base.dtype
            # Sample U and set S = 1, V = Id
            U = self[0].sample(distribution=distribution, init_=init_)
            S = torch.ones(
                *(self.tensorial_size + (self.n,)), device=device, dtype=dtype
            )
            V = torch.eye(self.n, device=device, dtype=dtype)
            if len(self.tensorial_size) > 0:
                V = V.repeat(*(self.tensorial_size + (1, 1)))

            if factorized:
                return U, S, V
            else:
                Vt = V.transpose(-2, -1)
                # Multiply the three of them, S as a diagonal matrix
                return U @ (S.unsqueeze(-1).expand_as(Vt) * Vt)

    def extra_repr(self):
        return _extra_repr(
            n=self.n,
            lam=self.lam,
            tensorial_size=self.tensorial_size,
            f=self.f,
            no_inv=self.inv is None,
        )



# Tests for the Stiefel manifold, grassmannian and SO(n)
from unittest import TestCase

from geotorch.so import SO
from geotorch.stiefel import Stiefel
from geotorch.grassmannian import Grassmannian
from geotorch.exceptions import NonSquareError, VectorError


class TestOrthogonal(TestCase):
    def test_constructor_stiefel(self):
        self._test_constructor(Stiefel)

    def test_constructors_grassmannian(self):
        self._test_constructor(Grassmannian)

    def _test_constructor(self, cls):
        with self.assertRaises(ValueError):
            cls(size=(3, 3), triv="wrong")

        with self.assertRaises(ValueError):
            SO(size=(3, 3), triv="wrong")

        # Try a custom trivialization (it should break in the forward)
        try:
            cls(size=(3, 3), triv=lambda: 3)
        except ValueError:
            self.fail("{} raised ValueError unexpectedly!".format(cls))

        # Try to instantiate it in a vector rather than a matrix
        with self.assertRaises(VectorError):
            cls(size=(7,))

        with self.assertRaises(VectorError):
            SO(size=(7,))

        # Try to instantiate it in an on-square matrix
        with self.assertRaises(NonSquareError):
            SO(size=(7, 3, 2))



import torch.nn as nn


class ProductManifold(nn.ModuleList):
    def __init__(self, manifolds):
        r"""
        Product manifold :math:`M_1 \times \dots \times M_k`. It can be indexed like a
        regular Python list.

        .. note::

            This is an abstract manifold. It may be used by composing it on the
            left and the right by an appropriate linear immersion / submersion.
            See for example the implementation in :class:`~geotorch.LowRank`

        Args:
            manifolds (iterable): An iterable of manifolds
        """
        super().__init__(manifolds)

    def forward(self, Xs):
        return tuple(mani(X) for mani, X in zip(self, Xs))

    def right_inverse(self, Xs, check_in_manifold=True):
        return tuple(
            mani.right_inverse(X, check_in_manifold) for mani, X in zip(self, Xs)
        )



import math
import torch
from torch import nn

try:
    from torch.linalg import qr
except ImportError:
    from torch import qr

from .utils import _extra_repr
from .skew import Skew

try:
    from torch import matrix_exp as expm
except ImportError:
    from .linalg.expm import expm
from .exceptions import NonSquareError, VectorError, InManifoldError


def _has_orthonormal_columns(X, eps):
    k = X.size(-1)
    Id = torch.eye(k, dtype=X.dtype, device=X.device)
    if X.dim() > 2:
        Id = Id.repeat(*(X.size()[:-2] + (1, 1)))
    return torch.allclose(X.transpose(-2, -1) @ X, Id, atol=eps)


def cayley_map(X):
    n = X.size(-1)
    Id = torch.eye(n, dtype=X.dtype, device=X.device)
    if X.ndimension() > 2:
        Id = Id.expand_as(X)
    halfX = 0.5 * X  # To make it into a retraction so that (d\phi)_0 = Id
    return torch.solve(Id + halfX, Id - halfX).solution


class SO(nn.Module):
    trivializations = {"expm": expm, "cayley": cayley_map}

    def __init__(self, size, triv="expm", lower=True):
        r"""
        Manifold of square orthogonal matrices with positive determinant parametrized
        in terms of its Lie algebra, the skew-symmetric matrices.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            triv (str or callable): Optional.
                A map that maps skew-symmetric onto :math:`\operatorname{SO}(n)`
                surjectively. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
            lower (bool): Optional. Uses the lower triangular part of the matrix to
                parametrize the skew-symmetric matrices. Default: ``True``
        """
        super().__init__()
        n, tensorial_size = SO.parse_size(size)
        self.n = n
        self.tensorial_size = tensorial_size
        self.lower = lower
        self.triv = SO.parse_triv(triv)
        self.register_buffer(
            "base", torch.empty(*(self.tensorial_size + (self.n, self.n)))
        )
        uniform_init_(self.base)

    @classmethod
    def parse_size(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        n = max(size[-2:])
        k = min(size[-2:])
        if n != k:
            raise NonSquareError(cls.__name__, size)
        tensorial_size = size[:-2]
        return n, tensorial_size

    @staticmethod
    def parse_triv(triv):
        if triv in SO.trivializations.keys():
            return SO.trivializations[triv]
        elif callable(triv):
            return triv
        else:
            raise ValueError(
                "Argument triv was not recognized and is "
                "not callable. Should be one of {}. Found {}".format(
                    list(SO.trivializations.keys()), triv
                )
            )

    def forward(self, X):
        X = Skew.frame(X, self.lower)
        return self.base @ self.triv(X)

    def right_inverse(self, X, check_in_manifold=True):
        if check_in_manifold and not self.in_manifold(X):
            raise InManifoldError(X, self)
        with torch.no_grad():
            self.base.copy_(X)
        return torch.zeros_like(X)

    def in_manifold(self, X, in_so=False, eps=1e-4):
        r"""
        Checks that a matrix is in the manifold.

        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor): The matrix to be checked
            in_so (bool): Optional. Checks that the matrix is orthogonal and
                has positive determinant. Otherwise just orthogonality is checked.
                Default: ``False``
            eps (float): Optional. Tolerance to numerical errors.
                Default: ``1e-4``
        """
        if X.size() != self.base.size():
            return False
        is_orth = _has_orthonormal_columns(X, eps)
        X_in_correct_coset = not in_so or (X.det() > 0.0).all().item()
        return is_orth and X_in_correct_coset

    def sample(self, distribution="uniform", init_=None):
        r"""
        Returns a randomly sampled orthogonal matrix according to the specified
        ``distribution``. The options are:

            - ``"uniform"``: Samples a tensor distributed according to the Haar measure
              on :math:`\operatorname{SO}(n)`

            - ``"torus"``: Samples a block-diagonal skew-symmetric matrix.
              The blocks are of the form
              :math:`\begin{pmatrix} 0 & b \\ -b & 0\end{pmatrix}` where :math:`b` is
              distributed according to ``init_``. This matrix will be then projected
              onto :math:`\operatorname{SO}(n)` using ``self.triv``

        .. note

            The ``"torus"`` initialization is particularly useful in recurrent kernels
            of RNNs

        Args:
            distribution (string): Optional. One of ``["uniform", "torus"]``.
                    Default: ``"uniform"``
            init\_ (callable): Optional. To be used with the ``"torus"`` option.
                    A function that takes a tensor and fills it in place according
                    to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: :math:`\operatorname{Uniform}(-\pi, \pi)`
        """
        device = self.base.device
        dtype = self.base.dtype
        ret = torch.empty(
            *(self.tensorial_size + (self.n, self.n)), device=device, dtype=dtype
        )
        if distribution == "uniform":
            uniform_init_(ret)
        elif distribution == "torus":
            torus_init_(ret, init_, self.triv)
        else:
            raise ValueError(
                'The ditribution has to be one of ["uniform", "torus"]. '
                "Got {}".format(distribution)
            )
        return ret

    def extra_repr(self):
        return _extra_repr(n=self.n, tensorial_size=self.tensorial_size, triv=self.triv)


def uniform_init_(tensor):
    r"""Fills in the input ``tensor`` in place with an orthogonal matrix.
    If square, the matrix will have positive determinant.
    The tensor will be distributed according to the Haar measure.
    The input tensor must have at least 2 dimensions.
    For tensors with more than 2 dimensions the first dimensions are treated as
    batch dimensions.

    Args:
        tensor (torch.Tensor): a 2-dimensional tensor or a batch of them
    """
    # We re-implement torch.nn.init.orthogonal_, as their treatment of batches
    # is not in a per-matrix base
    if tensor.ndimension() < 2:
        raise ValueError(
            "Only tensors with 2 or more dimensions are supported. "
            "Got a tensor of shape {}".format(tuple(tensor.size()))
        )
    n, k = tensor.size()[-2:]
    transpose = n < k
    with torch.no_grad():
        x = torch.empty_like(tensor).normal_(0, 1)
        if transpose:
            x.transpose_(-2, -1)
        q, r = qr(x)

        # Make uniform (diag r >= 0)
        d = r.diagonal(dim1=-2, dim2=-1).sign()
        q *= d.unsqueeze(-2).expand_as(q)
        if transpose:
            q.transpose_(-2, -1)

        # Make them have positive determinant by multiplying the
        # first column by -1 (does not change the measure)
        if n == k:
            mask = (torch.det(q) > 0.0).float()
            mask[mask == 0.0] = -1.0
            mask = mask.unsqueeze(-1).unsqueeze(-1).expand_as(q)
            q[..., 0] *= mask[..., 0]
        tensor.copy_(q)
        return tensor


def torus_init_(tensor, init_=None, triv=expm):
    r"""Fills in the input ``tensor`` in place as a block-diagonal skew-symmetric matrix.
    The blocks are of the form
    :math:`\begin{pmatrix} 0 & b \\ -b & 0\end{pmatrix}` where :math:`b` is
    distributed according to ``init_``.
    This matrix is then projected onto the manifold using ``triv``.

    The input tensor must have at least 2 dimension. For tensors with more than 2 dimensions
    the first dimensions are treated as batch dimensions.

    Args:
        tensor (torch.Tensor): a 2-dimensional tensor
        init\_ (callable): Optional. A function that takes a tensor and fills
                it in place according to some distribution. See
                `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                Default: :math:`\operatorname{Uniform}(-\pi, \pi)`
        triv (callable): Optional. A function that maps skew-symmetric matrices
                to orthogonal matrices.
    """
    if tensor.ndimension() < 2 or tensor.size(-1) != tensor.size(-2):
        raise ValueError(
            "Only tensors with 2 or more dimensions which are square in "
            "the last two dimensions are supported. "
            "Got a tensor of shape {}".format(tuple(tensor.size()))
        )

    n = tensor.size(-2)
    tensorial_size = tensor.size()[:-2]

    # Non-zero elements that we are going to set on the diagonal
    n_diag = n // 2
    diag = tensor.new(tensorial_size + (n_diag,))
    if init_ is None:
        torch.nn.init.uniform_(diag, -math.pi, math.pi)
    else:
        init_(diag)

    with torch.no_grad():
        # First non-central diagonal
        diag_z = tensor.new_zeros(tensorial_size + (n - 1,))
        diag_z[..., ::2] = diag
        x = torch.diag_embed(diag_z, offset=-1)
        tensor.copy_(triv(x - x.transpose(-2, -1)))
    return tensor



import torch
from functools import partial

try:
    from torch.linalg import svd

    svd = partial(svd, full_matrices=False)
except ImportError:
    from torch import svd


from .product import ProductManifold
from .stiefel import Stiefel
from .reals import Rn
from .exceptions import VectorError, RankError, InManifoldError
from .utils import transpose, _extra_repr


class LowRank(ProductManifold):
    def __init__(self, size, rank, triv="expm"):
        r"""
        Variety of the matrices of rank :math:`r` or less.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            rank (int): Rank of the matrices.
                It has to be less or equal to
                :math:`\min(\texttt{size}[-1], \texttt{size}[-2])`
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`U` and :math:`V` in
                the SVD. It can be one of ``["expm", "cayley"]`` or a custom callable.
                Default: ``"expm"``
        """
        n, k, tensorial_size, transposed = LowRank.parse_size(size)
        if rank > min(n, k) or rank < 1:
            raise RankError(n, k, rank)
        super().__init__(LowRank.manifolds(n, k, rank, tensorial_size, triv))
        self.n = n
        self.k = k
        self.rank = rank
        self.tensorial_size = tensorial_size
        self.transposed = transposed

    @classmethod
    def parse_size(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        transposed = size[-2] < size[-1]
        n = max(size[-2:])
        k = min(size[-2:])
        tensorial_size = size[:-2]
        return n, k, tensorial_size, transposed

    @staticmethod
    def manifolds(n, k, rank, tensorial_size, triv):
        size_u = tensorial_size + (n, rank)
        size_s = tensorial_size + (rank,)
        size_v = tensorial_size + (k, rank)
        return Stiefel(size_u, triv), Rn(size_s), Stiefel(size_v, triv)

    def frame(self, X):
        U = X.tril(-1)[..., : self.rank]
        S = X.diagonal(dim1=-2, dim2=-1)[..., : self.rank]
        V = X.triu(1).transpose(-2, -1)[..., : self.rank]
        return U, S, V

    def submersion(self, U, S, V):
        Vt = V.transpose(-2, -1)
        # Multiply the three of them, S as a diagonal matrix
        return U @ (S.unsqueeze(-1).expand_as(Vt) * Vt)

    @transpose
    def forward(self, X):
        X = self.frame(X)
        U, S, V = super().forward(X)
        return self.submersion(U, S, V)

    def frame_inv(self, X1, X2, X3):
        with torch.no_grad():
            # X1 is lower-triangular
            # X2 is a vector
            # X3 is upper-triangular
            size = self.tensorial_size + (self.n, self.k)
            ret = torch.zeros(*size, dtype=X1.dtype, device=X1.device)
            ret[..., : self.rank] += X1
            ret[..., : self.rank, : self.rank] += torch.diag_embed(X2)
            ret[..., : self.rank, :] += X3.transpose(-2, -1)
        return ret

    def submersion_inv(self, X, check_in_manifold=True):
        if isinstance(X, torch.Tensor):
            U, S, V = svd(X)
            if check_in_manifold and not self.in_manifold_singular_values(S):
                raise InManifoldError(X, self)
        else:
            # We assume that we got he U S V factorized in a tuple / list
            U, S, V = X
            if check_in_manifold and not self.in_manifold_tuple(U, S, V):
                raise InManifoldError(X, self)
        return U[..., : self.rank], S[..., : self.rank], V[..., : self.rank]

    @transpose
    def right_inverse(self, X, check_in_manifold=True):
        USV = self.submersion_inv(X, check_in_manifold)
        X1, X2, X3 = super().right_inverse(USV, check_in_manifold=False)
        return self.frame_inv(X1, X2, X3)

    def in_manifold_singular_values(self, S, eps=1e-5):
        r"""
        Checks that an ordered vector of singular values is in the manifold.

        For tensors with more than 1 dimension the first dimensions are
        treated as batch dimensions.

        Args:
            S (torch.Tensor): Vector of singular values
            eps (float): Optional. Threshold at which the singular values are
                considered to be zero
                Default: ``1e-5``
        """
        if S.size(-1) <= self.rank:
            return True
        # We compute the \infty-norm of the remaining dimension
        D = S[..., self.rank :]
        infty_norm_err = D.abs().max(dim=-1).values
        return (infty_norm_err < eps).all()

    def in_manifold_tuple(self, U, S, V, eps=1e-5):
        return (
            self.in_manifold_singular_values(S, eps)
            and self[0].in_manifold(U)
            and self[1].in_manifold(S)
            and self[2].in_manifold(V)
        )

    def in_manifold(self, X, eps=1e-5):
        r"""
        Checks that a matrix is in the manifold. The matrix may be given
        factorized in a `3`-tuple :math:`(U, \Sigma, V)` of a matrix, vector,
        and matrix representing an SVD of the matrix.


        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor or tuple): The matrix to be checked or a tuple containing
                :math:`(U, \Sigma, V)` as returned by ``torch.linalg.svd`` or
                ``self.sample(factorized=True)``.
            eps (float): Optional. Threshold at which the singular values are
                    considered to be zero
                    Default: ``1e-5``
        """
        if isinstance(X, tuple):
            if len(X) == 3:
                return self.in_manifold_tuple(X[0], X[1], X[2])
            else:
                return False
        else:
            if X.size(-1) > X.size(-2):
                X = X.transpose(-2, -1)
            if X.size() != self.tensorial_size + (self.n, self.k):
                return False
            try:
                S = torch.linalg.svdvals(X)
            except AttributeError:
                S = svd(X).S
            return self.in_manifold_singular_values(S, eps)

    def project(self, X, factorized=True):
        r"""
        Project a matrix onto the manifold.

        If ``factorized==True``, it returns a tuple containing the SVD decomposition of
        the matrix.

        Args:
            X (torch.Tensor): Matrix to be projected onto the manifold
            factorized (bool): Optional. Return an SVD decomposition of the
                    sampled matrix as a tuple :math:`(U, \Sigma, V)`.
                    Using ``factorized=True`` is more efficient when the result is
                    used to initialize a parametrized tensor.
                    Default: ``True``
        """
        U, S, V = svd(X)
        U, S, V = U[..., : self.rank], S[..., : self.rank], V[..., : self.rank]
        if factorized:
            return U, S, V
        else:
            X = self.submersion(U, S, V)
            if self.transposed:
                X = X.transpose(-2, -1)
            return X

    def sample(self, init_=torch.nn.init.xavier_normal_, factorized=True):
        r"""
        Returns a randomly sampled matrix on the manifold by sampling a matrix according
        to ``init_`` and projecting it onto the manifold.

        The output of this method can be used to initialize a parametrized tensor
        that has been parametrized with this or any other manifold as::

            >>> layer = nn.Linear(20, 20)
            >>> M = LowRank(layer.weight.size(), rank=6)
            >>> geotorch.register_parametrization(layer, "weight", M)
            >>> layer.weight = M.sample()

        Args:
            init\_ (callable): Optional. A function that takes a tensor and fills it
                    in place according to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: ``torch.nn.init.xavier_normal_``
            factorized (bool): Optional. Return an SVD decomposition of the
                    sampled matrix as a tuple :math:`(U, \Sigma, V)`.
                    Using ``factorized=True`` is more efficient when the result is
                    used to initialize a parametrized tensor.
                    Default: ``True``
        """
        with torch.no_grad():
            device = self[0].base.device
            dtype = self[0].base.dtype
            X = torch.empty(
                *(self.tensorial_size + (self.n, self.k)), device=device, dtype=dtype
            )
            init_(X)
            U, S, V = svd(X)
            U, S, V = U[..., : self.rank], S[..., : self.rank], V[..., : self.rank]
            if factorized:
                return U, S, V
            else:
                Vt = V.transpose(-2, -1)
                # Multiply the three of them, S as a diagonal matrix
                X = U @ (S.unsqueeze(-1).expand_as(Vt) * Vt)
                if self.transposed:
                    X = X.transpose(-2, -1)
                return X

    def extra_repr(self):
        return _extra_repr(
            n=self.n,
            k=self.k,
            rank=self.rank,
            tensorial_size=self.tensorial_size,
            transposed=self.transposed,
        )



import torch
from torch.nn.modules.container import ModuleList, ModuleDict, Module
from torch.nn.parameter import Parameter
from torch import Tensor
from typing import Union, Optional, Iterable, Dict, Tuple
from contextlib import contextmanager


_cache_enabled = 0
_cache: Dict[Tuple[int, str], Optional[Tensor]] = {}


@contextmanager
def cached():
    r"""Context manager that enables the caching system within parametrizations
    registered with :func:`register_parametrization`.

    The value of the parametrized objects is computed and cached the first time
    they are required when this context manager is active. The cached values are
    discarded when leaving the context manager.

    This is useful when using a parametrized parameter more than once in the forward pass.
    An example of this is when parametrizing the recurrent kernel of an RNN or when
    sharing weights.

    The simplest way to activate the cache is by wrapping the forward pass of the neural network

    .. code-block:: python

        import torch.nn.utils.parametrize as P
        ...
        with P.cached():
            output = model(inputs)

    in training and evaluation. One may also wrap the parts of the modules that use
    several times the parametrized tensors. For example, the loop of an RNN with a
    parametrized recurrent kernel:

    .. code-block:: python

        with P.cached():
            for x in xs:
                out_rnn = self.rnn_cell(x, out_rnn)
    """
    global _cache
    global _cache_enabled
    _cache_enabled += 1
    try:
        yield
    finally:
        _cache_enabled -= 1
        if not _cache_enabled:
            _cache = {}


class ParametrizationList(ModuleList):
    r"""A sequential container that holds and manages the ``original`` parameter or buffer of
    a parametrized :class:`torch.nn.Module`. It is the type of
    ``module.parametrizations[tensor_name]`` when ``module[tensor_name]`` has been parametrized
    with :func:`register_parametrization`.

    .. note ::
        This class is used internally by :func:`register_parametrization`. It is documented
        here for completeness. It should not be instantiated by the user.

    Args:
        modules (iterable): an iterable of modules representing the parametrizations
        original (Parameter or Tensor): parameter or buffer that is parametrized
    """
    original: Tensor

    def __init__(
        self, modules: Iterable[Module], original: Union[Tensor, Parameter]
    ) -> None:
        super().__init__(modules)
        if isinstance(original, Parameter):
            self.register_parameter("original", original)
        else:
            self.register_buffer("original", original)

    def set_original_(self, value: Tensor) -> None:
        r"""This method is called when assigning to a parametrized tensor.

        It calls the methods ``right_inverse`` (see :func:`register_parametrization`)
        of the parametrizations in the inverse order that they have been registered.
        Then, it assigns the result to ``self.original``.

        Args:
            value (Tensor): Value to which initialize the module

        Raises:
            RuntimeError: if any of the parametrizations do not implement a ``right_inverse`` method
        """
        with torch.no_grad():
            # See https://github.com/pytorch/pytorch/issues/53103
            for module in reversed(self):  # type: ignore
                if hasattr(module, "right_inverse"):
                    value = module.right_inverse(value)
                else:
                    raise RuntimeError(
                        "The parametrization '{}' does not implement a 'right_inverse' method. "
                        "Assigning to a parametrized tensor is only possible when all the parametrizations "
                        "implement a 'right_inverse' method.".format(
                            module.__class__.__name__
                        )
                    )
            self.original.copy_(value)

    def forward(self) -> Tensor:
        x = self.original
        for module in self:
            x = module(x)
        if x.size() != self.original.size():
            raise RuntimeError(
                "The parametrization may not change the size of the parametrized tensor. "
                "Size of original tensor: {} "
                "Size of parametrized tensor: {}".format(self.original.size(), x.size())
            )
        return x


def _inject_new_class(module: Module) -> None:
    r"""Sets up the parametrization mechanism used by parametrizations.

    This works by substituting the class of the module by a class
    that extends it to be able to inject a property

    Args:
        module (nn.Module): module into which to inject the property
    """
    cls = module.__class__

    def getstate(self):
        raise RuntimeError(
            "Serialization of parametrized modules is only "
            "supported through state_dict(). See:\n"
            "https://pytorch.org/tutorials/beginner/saving_loading_models.html"
            "#saving-loading-a-general-checkpoint-for-inference-and-or-resuming-training"
        )

    param_cls = type(
        "Parametrized{}".format(cls.__name__),
        (cls,),
        {
            "__getstate__": getstate,
        },
    )

    module.__class__ = param_cls


def _inject_property(module: Module, tensor_name: str) -> None:
    r"""Injects a property into module[tensor_name].

    It assumes that the class in the module has already been modified from its
    original one using _inject_new_class and that the tensor under :attr:`tensor_name`
    has already been moved out

    Args:
        module (nn.Module): module into which to inject the property
        tensor_name (str): name of the name of the property to create
    """
    # We check the precondition.
    # This should never fire if register_parametrization is correctly implemented
    assert not hasattr(module, tensor_name)

    def get_parametrized(self) -> Tensor:
        global _cache

        parametrization = self.parametrizations[tensor_name]
        if _cache_enabled:
            key = (id(module), tensor_name)
            tensor = _cache.get(key)
            if tensor is None:
                tensor = parametrization()
                _cache[key] = tensor
            return tensor
        else:
            # If caching is not active, this function just evaluates the parametrization
            return parametrization()

    def set_original(self, value: Tensor) -> None:
        self.parametrizations[tensor_name].set_original_(value)

    setattr(module.__class__, tensor_name, property(get_parametrized, set_original))


def register_parametrization(
    module: Module, tensor_name: str, parametrization: Module
) -> Module:
    r"""Adds a parametrization to a tensor in a module.

    Assume that ``tensor_name="weight"`` for simplicity. When accessing ``module.weight``,
    the module will return the parametrized version ``parametrization(module.weight)``.
    If the original tensor requires a gradient, the backward pass will differentiate
    through the :attr:`parametrization`, and the optimizer will update the tensor accordingly.

    The first time that a module registers a parametrization, this function will add an attribute
    ``parametrizations`` to the module of type :class:`~ParametrizationList`.

    The list of parametrizations on a tensor will be accessible under
    ``module.parametrizations.weight``.

    The original tensor will be accessible under
    ``module.parametrizations.weight.original``.

    Parametrizations may be concatenated by registering several parametrizations
    on the same attribute.

    Parametrized parameters and buffers have an inbuilt caching system that can be activated
    using the context manager :func:`cached`.

    A :attr:`parametrization` may optionally implement a method with signature

    .. code-block:: python

        def right_inverse(self, X: Tensor) -> Tensor

    If :attr:`parametrization` implements this method, it will be possible to assign
    to the parametrized tensor. This may be used to initialize the tensor, as shown in the example.

    In most situations, ``right_inverse`` will be a function such that
    ``forward(right_inverse(X)) == X`` (see
    `right inverse <https://en.wikipedia.org/wiki/Inverse_function#Right_inverses>`_).
    Sometimes, when the parametrization is not surjective, it may be reasonable
    to relax this, as shown in the example below.

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (str): name of the parameter or buffer on which to register
            the parametrization
        parametrization (nn.Module): the parametrization to register

    Returns:
        Module: module

    Raises:
        ValueError: if the module does not have a parameter or a buffer named :attr:`tensor_name`

    Examples:
        >>> import torch
        >>> import torch.nn.utils.parametrize as P
        >>>
        >>> class Symmetric(torch.nn.Module):
        >>>     def forward(self, X):
        >>>         return X.triu() + X.triu(1).T  # Return a symmetric matrix
        >>>
        >>>     def right_inverse(self, A):
        >>>         return A.triu()
        >>>
        >>> m = torch.nn.Linear(5, 5)
        >>> P.register_parametrization(m, "weight", Symmetric())
        >>> print(torch.allclose(m.weight, m.weight.T))  # m.weight is now symmetric
        True
        >>> A = torch.rand(5, 5)
        >>> A = A + A.T   # A is now symmetric
        >>> m.weight = A  # Initialize the weight to be the symmetric matrix A
        >>> print(torch.allclose(m.weight, A))
        True
    """
    if is_parametrized(module, tensor_name):
        # Just add the new parametrization to the parametrization list
        module.parametrizations[tensor_name].append(parametrization)  # type: ignore
    elif tensor_name in module._buffers or tensor_name in module._parameters:
        # Set the parametrization mechanism
        # Fetch the original buffer or parameter
        original = getattr(module, tensor_name)
        # Delete the previous parameter or buffer
        delattr(module, tensor_name)
        # If this is the first parametrization registered on the module,
        # we prepare the module to inject the property
        if not is_parametrized(module):
            # Change the class
            _inject_new_class(module)
            # Inject the a ``ModuleDict`` into the instance under module.parametrizations
            module.parametrizations = ModuleDict()
        # Add a property into the class
        _inject_property(module, tensor_name)
        # Add a ParametrizationList
        module.parametrizations[tensor_name] = ParametrizationList(  # type: ignore
            [parametrization], original
        )
    else:
        raise ValueError(
            "Module '{}' does not have a parameter, a buffer, or a "
            "parametrized element with name '{}'".format(module, tensor_name)
        )
    return module


def is_parametrized(module: Module, tensor_name: Optional[str] = None) -> bool:
    r"""Returns ``True`` if module has an active parametrization.

    If the argument :attr:`tensor_name` is specified, returns ``True`` if
    ``module[tensor_name]`` is parametrized.

    Args:
        module (nn.Module): module to query
        name (str, optional): attribute in the module to query
            Default: ``None``
    """
    parametrizations = getattr(module, "parametrizations", None)
    if parametrizations is None or not isinstance(parametrizations, ModuleDict):
        return False
    if tensor_name is None:
        # Check that there is at least one parametrized buffer or Parameter
        return len(parametrizations) > 0
    else:
        return tensor_name in parametrizations


def remove_parametrizations(
    module: Module, tensor_name: str, leave_parametrized: bool = True
) -> Module:
    r"""Removes the parametrizations on a tensor in a module.

    - If ``leave_parametrized=True``, ``module[tensor_name]`` will be set to
      its current output. In this case, the parametrization shall not change the ``dtype``
      of the tensor.
    - If ``leave_parametrized=False``, ``module[tensor_name]`` will be set to
      the unparametrised tensor in ``module.parametrizations[tensor_name].original``.

    Args:
        module (nn.Module): module from which remove the parametrization
        tensor_name (str): name of the parametrization to be removed
        leave_parametrized (bool, optional): leave the attribute :attr:`tensor_name` parametrized.
            Default: ``True``

    Returns:
        Module: module

    Raises:
        ValueError: if ``module[tensor_name]`` is not parametrized
        ValueError: if ``leave_parametrized=True`` and the parametrization changes the size or dtype
            of the tensor
    """

    if not is_parametrized(module, tensor_name):
        raise ValueError(
            "Module {} does not have a parametrization on {}".format(
                module, tensor_name
            )
        )

    # Fetch the original tensor
    original = module.parametrizations[tensor_name].original  # type: ignore
    if leave_parametrized:
        t = getattr(module, tensor_name)
        # If they have the same dtype, we reuse the original tensor.
        # We do this so that the parameter does not to change the id()
        # This way the user does not need to update the optimizer
        if t.dtype == original.dtype:
            with torch.no_grad():
                original.set_(t)
        else:
            raise ValueError(
                "The parametrization changes the dtype of the tensor from {} to {}. "
                "It is not supported to leave the tensor parametrized (`leave_parametrized=True`) "
                "in this case.".format(original.dtype, t.dtype)
            )
    # Delete the property that manages the parametrization
    delattr(module.__class__, tensor_name)
    # Delete the ParametrizationList
    del module.parametrizations[tensor_name]  # type: ignore

    # Restore the parameter / buffer into the main class
    if isinstance(original, Parameter):
        module.register_parameter(tensor_name, original)
    else:
        module.register_buffer(tensor_name, original)

    # Roll back the parametrized class if no other buffer or parameter
    # is currently parametrized in this class
    if not is_parametrized(module):
        delattr(module, "parametrizations")
        # Restore class
        orig_cls = module.__class__.__bases__[0]
        module.__class__ = orig_cls
    return module



from .pssdfixedrank import PSSDFixedRank
from .exceptions import VectorError, NonSquareError
from .utils import _extra_repr


class PSD(PSSDFixedRank):
    def __init__(self, size, f="softplus", triv="expm"):
        r"""
        Manifold of symmetric positive definite matrices

        Args:
            size (torch.size): Size of the tensor to be parametrized
            f (str or callable or pair of callables): Optional. Either:

                - ``"softplus"``

                - A callable that maps real numbers to the interval :math:`(0, \infty)`

                - A pair of callables such that the first maps the real numbers to
                  :math:`(0, \infty)` and the second is a (right) inverse of the first

                Default: ``"softplus"``
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`Q` in the eigenvalue
                decomposition. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size, PSD.rank(size), f, triv=triv)

    @classmethod
    def rank(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        n, k = size[-2:]
        if n != k:
            raise NonSquareError(cls.__name__, size)
        return n

    def extra_repr(self):
        return _extra_repr(
            n=self.n,
            tensorial_size=self.tensorial_size,
            f=self.f,
            no_inv=self.inv is None,
        )



import torch
from .lowrank import LowRank
from .exceptions import InverseError


def softplus_epsilon(x, epsilon=1e-6):
    return torch.nn.functional.softplus(x) + epsilon


def inv_softplus_epsilon(x, epsilon=1e-6):
    y = x - epsilon
    return torch.where(y > 20, y, y.expm1().log())


class FixedRank(LowRank):
    fs = {"softplus": (softplus_epsilon, inv_softplus_epsilon)}

    def __init__(self, size, rank, f="softplus", triv="expm"):
        r"""
        Manifold of non-square matrices of rank equal to ``rank``

        Args:
            size (torch.size): Size of the tensor to be parametrized
            rank (int): Rank of the matrices.
                It has to be less or equal to
                :math:`\min(\texttt{size}[-1], \texttt{size}[-2])`
            f (str or callable or pair of callables): Optional. Either:

                - ``"softplus"``

                - A callable that maps real numbers to the interval :math:`(0, \infty)`

                - A pair of callables such that the first maps the real numbers onto
                  :math:`(0, \infty)` and the second is a (right) inverse of the first

                Default: ``"softplus"``
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`U` and :math:`V` in
                the SVD. It can be one of ``["expm", "cayley"]`` or a custom callable.
                Default: ``"expm"``
        """
        super().__init__(size, rank, triv=triv)
        f, inv = FixedRank.parse_f(f)
        self.f = f
        self.inv = inv

    @staticmethod
    def parse_f(f):
        if f in FixedRank.fs.keys():
            return FixedRank.fs[f]
        elif callable(f):
            return f, None
        elif isinstance(f, tuple) and callable(f[0]) and callable(f[1]):
            return f
        else:
            raise ValueError(
                "Argument f was not recognized and is "
                "not callable or a pair of callables. "
                "Should be one of {}. Found {}".format(list(FixedRank.fs.keys()), f)
            )

    def submersion(self, U, S, V):
        return super().submersion(U, self.f(S), V)

    def submersion_inv(self, X, check_in_manifold=True):
        U, S, V = super().submersion_inv(X, check_in_manifold)
        if self.inv is None:
            raise InverseError(self)
        return U, self.inv(S), V

    def in_manifold_singular_values(self, S, eps=1e-5):
        r"""
        Checks that a vector of singular values is in the manifold.

        For tensors with more than 1 dimension the first dimensions are
        treated as batch dimensions.

        Args:
            S (torch.Tensor): Vector of singular values
            eps (float): Optional. Threshold at which the singular values are
                considered to be zero
                Default: ``1e-5``
        """
        if not super().in_manifold_singular_values(S, eps):
            return False
        # We compute the \infty-norm of the eigenvalues
        D = S[..., : self.rank]
        infty_norm = D.abs().max(dim=-1).values
        return (infty_norm > eps).all().item()

    def sample(self, init_=torch.nn.init.xavier_normal_, factorized=True, eps=5e-6):
        r"""
        Returns a randomly sampled matrix on the manifold by sampling a matrix according
        to ``init_`` and projecting it onto the manifold.

        If the sampled matrix has more than `self.rank` small singular values, the
        smallest ones are clamped to be at least ``eps`` in absolute value.

        The output of this method can be used to initialize a parametrized tensor
        that has been parametrized with this or any other manifold as::

            >>> layer = nn.Linear(20, 20)
            >>> M = FixedRank(layer.weight.size(), rank=6)
            >>> geotorch.register_parametrization(layer, "weight", M)
            >>> layer.weight = M.sample()

        Args:
            init\_ (callable): Optional. A function that takes a tensor and fills it
                    in place according to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: ``torch.nn.init.xavier_normal_``
            factorized (bool): Optional. Return an SVD decomposition of the
                    sampled matrix as a tuple :math:`(U, \Sigma, V)`.
                    Using ``factorized=True`` is more efficient when the result is
                    used to initialize a parametrized tensor.
                    Default: ``True``
            eps (float): Optional. Minimum singular value of the sampled matrix.
                    Default: ``5e-6``
        """
        U, S, V = super().sample(factorized=True, init_=init_)
        with torch.no_grad():
            # S >= 0, as given by torch.linalg.eigvalsh()
            S[S < eps] = eps
        if factorized:
            return U, S, V
        else:
            Vt = V.transpose(-2, -1)
            # Multiply the three of them, S as a diagonal matrix
            X = U @ (S.unsqueeze(-1).expand_as(Vt) * Vt)
            if self.transposed:
                X = X.transpose(-2, -1)
            return X



import torch
from torch import nn
from .exceptions import VectorError, NonSquareError


class Skew(nn.Module):
    def __init__(self, lower=True):
        r"""
        Vector space of skew-symmetric matrices, parametrized in terms of
        the upper or lower triangular part of a matrix.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            lower (bool): Optional. Uses the lower triangular part of the matrix
                to parametrize the matrix. Default: ``True``
        """
        super().__init__()
        self.lower = lower

    @staticmethod
    def frame(X, lower):
        if lower:
            X = X.tril(-1)
        else:
            X = X.triu(1)
        return X - X.transpose(-2, -1)

    def forward(self, X):
        if len(X.size()) < 2:
            raise VectorError(type(self).__name__, X.size())
        if X.size(-2) != X.size(-1):
            raise NonSquareError(type(self).__name__, X.size())
        return self.frame(X, self.lower)

    @staticmethod
    def in_manifold(X):
        return (
            X.dim() >= 2
            and X.size(-2) == X.size(-1)
            and torch.allclose(X, -X.transpose(-2, -1))
        )



"""
Basic example of usage of GeoTorch

Implements a constrained RNN to learn a synthetic regression problem which asks to recall
some inputs and output them. From a dictionary of 9 numbers, the input-output looks like follows:

Input:   14221----------:----
Output:  ---------------14221

This class should converge to 0% error. When at 0% error, sometimes there are some instabilities.

The GeoTorch code happens in `ExpRNNCell.__init__`, `ExpRNNCell.reset_parameters` and line 107.
The rest of the code is normal PyTorch.
Lines 146-167 shows how to assign different learning rates to parametrized weights.

This file also implements in lines 152 and 180 Riemannian Gradient Descent (RGD). As shown, RGD
dynamics account for using SGD as the optimizer calling `update_basis()` after every optization step.
"""

import torch
from torch import nn
import torch.nn.functional as F

import geotorch

batch_size = 128
hidden_size = 190
iterations = 4000  # Training iterations
L = 1000  # Length of sequence before asking to remember
S = 10  # Length of sequence to remember
alphabet_size = 8
lr = 1e-3
lr_orth = 2e-4
device = torch.device("cuda")
# When RGD == True we perform Riemannian gradient descent
# This is to demonstrate how one may implement RGD with
# just one extra line of code.
# RGD does not perform very well in these problems though.
RGD = False
if RGD:
    print(
        "Optimizing using RGD. The perfomance will be _much_ worse than with Adam or RMSprop."
    )


class modrelu(nn.Module):
    def __init__(self, features):
        super(modrelu, self).__init__()
        self.features = features
        self.b = nn.Parameter(torch.Tensor(self.features))
        self.reset_parameters()

    def reset_parameters(self):
        self.b.data.uniform_(-0.01, 0.01)

    def forward(self, inputs):
        norm = torch.abs(inputs)
        biased_norm = norm + self.b
        magnitude = nn.functional.relu(biased_norm)
        phase = torch.sign(inputs)

        return phase * magnitude


class ExpRNNCell(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(ExpRNNCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.recurrent_kernel = nn.Linear(hidden_size, hidden_size, bias=False)
        self.input_kernel = nn.Linear(input_size, hidden_size)
        self.nonlinearity = modrelu(hidden_size)

        # Make recurrent_kernel orthogonal
        geotorch.orthogonal(self.recurrent_kernel, "weight")

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_normal_(self.input_kernel.weight.data, nonlinearity="relu")
        # The manifold class is under `layer.parametrizations.tensor_name[0]`
        M = self.recurrent_kernel.parametrizations.weight[0]
        # Every manifold has a convenience sample method, but you can use your own initializer
        self.recurrent_kernel.weight = M.sample("torus")

    def default_hidden(self, input_):
        return input_.new_zeros(input_.size(0), self.hidden_size, requires_grad=False)

    def forward(self, input_, hidden):
        input_ = self.input_kernel(input_)
        hidden = self.recurrent_kernel(hidden)
        out = input_ + hidden
        return self.nonlinearity(out)


class Model(nn.Module):
    def __init__(self, alphabet_size, hidden_size):
        super(Model, self).__init__()
        self.hidden_size = hidden_size
        self.rnn = ExpRNNCell(alphabet_size + 2, hidden_size)
        self.lin = nn.Linear(hidden_size, alphabet_size + 1)
        self.loss_func = nn.CrossEntropyLoss()
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_normal_(self.lin.weight.data, nonlinearity="relu")
        nn.init.constant_(self.lin.bias.data, 0)

    def forward(self, inputs):
        out_rnn = self.rnn.default_hidden(inputs[:, 0, ...])
        outputs = []
        with geotorch.parametrize.cached():
            for input in torch.unbind(inputs, dim=1):
                out_rnn = self.rnn(input, out_rnn)
                outputs.append(self.lin(out_rnn))
        return torch.stack(outputs, dim=1)

    def loss(self, logits, y):
        return self.loss_func(logits.view(-1, 9), y.view(-1))

    def accuracy(self, logits, y):
        return torch.eq(torch.argmax(logits, dim=2), y).float().mean()


def copy_data(batch_size):
    # Generates some random synthetic data
    # Example of input-output sequence
    # 14221----------:----
    # ---------------14221
    # Numbers go from 1 to 8
    # We generate S of them and we have to recall them
    # L is the waiting between the last number and the
    # signal to start outputting the numbers
    # We codify `-` as a 0 and `:` as a 9.

    seq = torch.randint(
        1, alphabet_size + 1, (batch_size, S), dtype=torch.long, device=device
    )
    zeros1 = torch.zeros((batch_size, L), dtype=torch.long, device=device)
    zeros2 = torch.zeros((batch_size, S - 1), dtype=torch.long, device=device)
    zeros3 = torch.zeros((batch_size, S + L), dtype=torch.long, device=device)
    marker = torch.full(
        (batch_size, 1), alphabet_size + 1, dtype=torch.long, device=device
    )

    x = torch.cat([seq, zeros1, marker, zeros2], dim=1)
    y = torch.cat([zeros3, seq], dim=1)

    return x, y


def main():
    model = Model(alphabet_size, hidden_size).to(device)

    p_orth = model.rnn.recurrent_kernel
    orth_params = p_orth.parameters()
    non_orth_params = (
        p for p in model.parameters() if p not in set(p_orth.parameters())
    )

    if RGD:
        # Implement Stochstic Riemannian Gradient Descent via SGD
        optim = torch.optim.SGD(
            [{"params": non_orth_params}, {"params": orth_params, "lr": lr_orth}], lr=lr
        )
    else:
        # These recurrent models benefit of slightly larger mixing constants
        # on the adaptive term. They also work with beta_2 = 0.999, but they
        # give better results with beta_2 \in [0.9, 0.99]
        optim = torch.optim.Adam(
            [
                {"params": non_orth_params},
                {"params": orth_params, "lr": lr_orth, "betas": (0.9, 0.95)},
            ],
            lr=lr,
        )

    model.train()
    for step in range(iterations):
        batch_x, batch_y = copy_data(batch_size)
        x_onehot = F.one_hot(batch_x, num_classes=alphabet_size + 2).float()
        logits = model(x_onehot)
        loss = model.loss(logits, batch_y)

        optim.zero_grad()
        loss.backward()
        optim.step()

        if RGD:
            # Updating the base after every step and using SGD gives us
            # Riemannian Gradient Descent. More on this in Section 5
            # https://arxiv.org/abs/1909.09501
            geotorch.update_base(model.rnn.recurrent_kernel, "weight")

        with torch.no_grad():
            accuracy = model.accuracy(logits, batch_y)

        print("Iter {} Loss: {:.6f}, Accuracy: {:.5f}".format(step, loss, accuracy))

    # The evaluation in this model is not quite necessary, as we do not repeat any
    # element of the training batch


if __name__ == "__main__":
    main()



import torch
import geotorch.parametrize as P

from .symmetric import Symmetric
from .skew import Skew
from .sphere import Sphere, SphereEmbedded
from .stiefel import Stiefel
from .grassmannian import Grassmannian
from .almostorthogonal import AlmostOrthogonal
from .lowrank import LowRank
from .fixedrank import FixedRank
from .glp import GLp
from .psd import PSD
from .pssd import PSSD
from .pssdlowrank import PSSDLowRank
from .pssdfixedrank import PSSDFixedRank


def _register_manifold(module, tensor_name, cls, *args):
    tensor = getattr(module, tensor_name)
    M = cls(tensor.size(), *args).to(device=tensor.device, dtype=tensor.dtype)
    P.register_parametrization(module, tensor_name, M)

    # Initialize without checking in manifold
    X = M.sample()
    param_list = module.parametrizations[tensor_name]
    with torch.no_grad():
        for m in reversed(param_list):
            X = m.right_inverse(X, check_in_manifold=False)
        param_list.original.copy_(X)

    return module


def symmetric(module, tensor_name="weight", lower=True):
    r"""Adds a symmetric parametrization to the matrix ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the parametrized
    version :math:`X` so that :math:`X^\intercal = X`.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(30, 30)
        >>> geotorch.symmetric(layer, "weight")
        >>> torch.allclose(layer.weight, layer.weight.T)
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        lower (bool): Optional. Uses the lower triangular part of the matrix to
            parametrize the matrix. Default: ``True``
    """
    P.register_parametrization(module, tensor_name, Symmetric(lower))
    return module


def skew(module, tensor_name="weight", lower=True):
    r"""Adds a skew-symmetric parametrization to the matrix ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the parametrized
    version :math:`X` so that :math:`X^\intercal = -X`.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(30, 30)
        >>> geotorch.skew(layer, "weight")
        >>> torch.allclose(layer.weight, -layer.weight.T)
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        lower (bool): Optional. Uses the lower triangular part of the matrix to
            parametrize the matrix. Default: ``True``
    """
    P.register_parametrization(module, tensor_name, Skew(lower))
    return module


def sphere(module, tensor_name="weight", radius=1.0, embedded=False):
    r"""Adds a spherical parametrization to the vector (or tensor) ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the parametrized
    version :math:`v` so that :math:`\lVert v \rVert = 1`.

    If the tensor has more than one dimension, the parametrization will be
    applied to the last dimension.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.sphere(layer, "bias")
        >>> torch.norm(layer.bias)
        tensor(1.)
        >>> geotorch.sphere(layer, "weight")  # Make the columns unit norm
        >>> torch.allclose(torch.norm(layer.weight, dim=-1), torch.ones(30))
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        radius (float): Optional.
            Radius of the sphere. It has to be positive. Default: 1.
        embedded (bool): Optional.
            Chooses between the implementation of the sphere using the exponential
            map (``embedded=False``) and that using the projection from the ambient space (``embedded=True``)
            Default. ``True``
    """
    cls = SphereEmbedded if embedded else Sphere
    return _register_manifold(module, tensor_name, cls, radius)


def orthogonal(module, tensor_name="weight", triv="expm"):
    r"""Adds an orthogonal parametrization to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` so that :math:`X^\intercal X = \operatorname{I}`.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.orthogonal(layer, "weight")
        >>> torch.norm(layer.weight.T @ layer.weight - torch.eye(20,20))
        tensor(4.8488e-05)

        >>> layer = nn.Conv2d(20, 40, 3, 3)  # Make the kernels orthogonal
        >>> geotorch.orthogonal(layer, "weight")
        >>> torch.norm(layer.weight.transpose(-2, -1) @ layer.weight - torch.eye(3,3).repeat(40,20,1,1))
        tensor(1.2225e-05)

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        triv (str or callable): Optional.
            A map that maps a skew-symmetric matrix to an orthogonal matrix.
            It can be the exponential of matrices or the cayley transform passing
            ``["expm", "cayley"]`` or a custom callable.  Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, Stiefel, triv)


def almost_orthogonal(module, tensor_name="weight", lam=0.1, f="sin", triv="expm"):
    r"""Adds an almost orthogonal parametrization to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will have its singular values in
    the interval :math:`[1-\texttt{lam}, 1+\texttt{lam}]`

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.almost_orthogonal(layer, "weight", 0.5)
        >>> S = torch.linalg.svd(layer.weight).S
        >>> all(S >= 0.5 and S <= 1.5)
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        lam (float): Radius of the interval for the singular values. A float in the interval :math:`[0, 1]`. Default: ``0.1``
        f (str or callable or pair of callables): Optional. Either:

            - One of ``["scaled_sigmoid", "tanh", "sin"]``

            - A callable that maps real numbers to the interval :math:`[-1, 1]`

            - A pair of callables such that the first maps the real numbers to
              :math:`[-1, 1]` and the second is a (right) inverse of the first

            Default: ``"sin"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal matrices
            surjectively. This is used to optimize the :math:`U` and :math:`V` in the
            SVD. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, AlmostOrthogonal, lam, f, triv)


def grassmannian(module, tensor_name="weight", triv="expm"):
    r"""Adds an parametrization to the tensor ``module.tensor_name`` so that the
    result represents a subspace. If the initial matrix was of size :math:`n \times k`
    the parametrized matrix will represent a subspace of dimension :math:`k` of
    :math:`\mathbb{R}^n`.

    When accessing ``module.tensor_name``, the module will return the parametrized
    version :math:`X` so that :math:`X` represents :math:`k` orthogonal vectors of
    :math:`\mathbb{R}^n` that span the subspace. That is, the resulting matrix will
    be orthogonal, :math:`X^\intercal X = \operatorname{I}`.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    .. note::

        Even though this space resembles that generated by :func:`geotorch.orthogonal`,
        it is actually a subspace of that, as every subspace can be represented by many
        different basis of vectors that span it.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.grassmannian(layer, "weight")
        >>> torch.norm(layer.weight.t() @ layer.weight - torch.eye(20,20))
        tensor(1.8933e-05)

        >>> layer = nn.Conv2d(20, 40, 3, 3)  # Make the kernels represent subspaces
        >>> geotorch.grassmannian(layer, "weight")
        >>> torch.norm(layer.weight.transpose(-2, -1) @ layer.weight - torch.eye(3,3).repeat(40,20,1,1))
        tensor(8.3796-06)

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        triv (str or callable): Optional.
            A map that maps a skew-symmetric matrix to an orthogonal matrix.
            It can be the exponential of matrices or the cayley transform passing
            ``["expm", "cayley"]`` or a custom callable.  Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, Grassmannian, triv)


def low_rank(module, tensor_name, rank, triv="expm"):
    r"""Adds a low rank parametrization to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will have rank at most ``rank``.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.low_rank(layer, "weight", 4)
        >>> list(torch.linalg.svd(layer.weight).S > 1e-7).count(True) <= 4
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied
        rank (int): Rank of the matrix.
            It has to be less than the minimum of the two dimensions of the
            matrix
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal matrices
            surjectively. This is used to optimize the :math:`U` and :math:`V` in the
            SVD. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, LowRank, rank, triv)


def fixed_rank(module, tensor_name, rank, f="softplus", triv="expm"):
    r"""Adds a fixed rank parametrization to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will have rank equal to ``rank``.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 30)
        >>> geotorch.fixed_rank(layer, "weight", 5)
        >>> list(torch.linalg.svd(layer.weight).S > 1e-7).count(True)
        5

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied
        rank (int): Rank of the matrix.
            It has to be less than the minimum of the two dimensions of the
            matrix
        f (str or callable or pair of callables): Optional. Either:

            - ``"softplus"``

            - A callable that maps real numbers to the interval :math:`(0, \infty)`

            - A pair of callables such that the first maps the real numbers to
              :math:`(0, \infty)` and the second is a (right) inverse of the first

            Default: ``"softplus"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal matrices
            surjectively. This is used to optimize the :math:`U` and :math:`V` in the
            SVD. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, FixedRank, rank, f, triv)


def invertible(module, tensor_name="weight", f="softplus", triv="expm"):
    r"""Adds an invertibility constraint to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will have positive determinant and,
    in particular, it will be invertible.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 20)
        >>> geotorch.invertible(layer, "weight", 5)
        >>> torch.det(layer.weight) > 0.0
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        f (str or callable or pair of callables): Optional. Either:

            - ``"softplus"``

            - A callable that maps real numbers to the interval :math:`(0, \infty)`

            - A pair of callables such that the first maps the real numbers to
              :math:`(0, \infty)` and the second is a (right) inverse of the first

            Default: ``"softplus"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal matrices
            surjectively. This is used to optimize the :math:`U` and :math:`V` in the
            SVD. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, GLp, f, triv)


def positive_definite(module, tensor_name="weight", f="softplus", triv="expm"):
    r"""Adds a positive definiteness constraint to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will be symmetric and with positive
    eigenvalues

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 20)
        >>> geotorch.positive_definite(layer, "weight")
        >>> (torch.linalg.eigvalsh(layer.weight) > 0.0).all()
        tensor(True)

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        f (str or callable or pair of callables): Optional. Either:

            - ``"softplus"``

            - A callable that maps real numbers to the interval :math:`(0, \infty)`

            - A pair of callables such that the first maps the real numbers to
              :math:`(0, \infty)` and the second is a (right) inverse of the first

            Default: ``"softplus"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal
            matrices surjectively. This is used to optimize the :math:`Q` in the eigenvalue
            decomposition. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, PSD, f, triv)


def positive_semidefinite(module, tensor_name="weight", triv="expm"):
    r"""Adds a positive definiteness constraint to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will be symmetric and with
    non-negative eigenvalues

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 20)
        >>> geotorch.positive_semidefinite(layer, "weight")
        >>> L = torch.linalg.eigvalsh(layer.weight)
        >>> L[L.abs() < 1e-7] = 0.0  # Round errors
        >>> (L >= 0.0).all()
        tensor(True)

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied. Default: ``"weight"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal
            matrices surjectively. This is used to optimize the :math:`Q` in the eigenvalue
            decomposition. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, PSSD, triv)


def positive_semidefinite_low_rank(module, tensor_name, rank, triv="expm"):
    r"""Adds a positive definiteness constraint to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will be symmetric and with non-negative
    eigenvalues and at most ``rank`` of them non-zero.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 20)
        >>> geotorch.positive_semidefinite_low_rank(layer, "weight", 5)
        >>> L = torch.linalg.eigvalsh(layer.weight)
        >>> L[L.abs() < 1e-7] = 0.0  # Round errors
        >>> (L >= 0.0).all()
        tensor(True)
        >>> list(L > 0.0).count(True) <= 5
        True

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied
        rank (int): Rank of the matrix.
            It has to be less than the minimum of the two dimensions of the
            matrix
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal
            matrices surjectively. This is used to optimize the :math:`Q` in the eigenvalue
            decomposition. It can be one of ``["expm", "cayley"]`` or a custom
            callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, PSSDLowRank, rank, triv)


def positive_semidefinite_fixed_rank(
    module, tensor_name, rank, f="softplus", triv="expm"
):
    r"""Adds a positive definiteness constraint to the tensor ``module.tensor_name``.

    When accessing ``module.tensor_name``, the module will return the
    parametrized version :math:`X` which will be symmetric and with non-negative
    eigenvalues and exactly ``rank`` of them non-zero.

    If the tensor has more than two dimensions, the parametrization will be
    applied to the last two dimensions.

    Examples::

        >>> layer = nn.Linear(20, 20)
        >>> geotorch.positive_semidefinite_fixed_rank(layer, "weight", 5)
        >>> L = torch.linalg.eigvalsh(layer.weight)
        >>> L[L.abs() < 1e-7] = 0.0  # Round errors
        >>> (L >= 0.0).all()
        tensor(True)
        >>> list(L > 0.0).count(True)
        5

    Args:
        module (nn.Module): module on which to register the parametrization
        tensor_name (string): name of the parameter, buffer, or parametrization
            on which the parametrization will be applied
        rank (int): Rank of the matrix.
            It has to be less than the minimum of the two dimensions of the
            matrix
        f (str or callable or pair of callables): Optional. Either:

            - ``"softplus"``

            - A callable that maps real numbers to the interval :math:`(0, \infty)`

            - A pair of callables such that the first maps the real numbers to
              :math:`(0, \infty)` and the second is a (right) inverse of the first

            Default: ``"softplus"``
        triv (str or callable): Optional.
            A map that maps skew-symmetric matrices onto the orthogonal
            matrices surjectively. This is used to optimize the :math:`Q` in the
            eigenvalue decomposition. It can be one of ``["expm", "cayley"]`` or
            a custom callable. Default: ``"expm"``
    """
    return _register_manifold(module, tensor_name, PSSDFixedRank, rank, f, triv)



"""
In this program we show how to use GeoTorch to compute the maximum eigenvalue
of a symmetric matrix via the Rayleigh quotient, restricting the optimisation
problem to the Sphere
"""
import torch

try:
    from torch.linalg import eigvalsh
except ImportError:
    from torch import symeig

    def eigvalsh(X):
        return symeig(X, eigenvectors=False).eigenvalues


from torch import nn
import geotorch

N = 1000  # matrix size
LR = 1.0 / N  # step-size.
# Obs. If the distribution of the matrix is changed, this parameter should be tuned


class Model(nn.Module):
    def __init__(self, n):
        super().__init__()
        self.x = nn.Parameter(torch.rand(n))
        geotorch.sphere(self, "x")
        Sphere = self.parametrizations.x[0]
        self.x = Sphere.sample()

    def forward(self, A):
        x = self.x
        return x.T @ A @ x


# Generate matrix
A = torch.rand(N, N)  # Uniform on [0, 1)
A = 0.5 * (A + A.T)

# Compare against diagonalization (eigenvalues are returend in ascending order)
max_eigenvalue = eigvalsh(A)[-1]
print("Max eigenvalue: {:10.5f}".format(max_eigenvalue))

# Instantiate model and optimiser
model = Model(N)
optim = torch.optim.SGD(model.parameters(), lr=LR)

eigenvalue = float("inf")
i = 0
while (eigenvalue - max_eigenvalue).abs() > 1e-3:
    eigenvalue = model(A)

    optim.zero_grad()
    (-eigenvalue).backward()
    optim.step()
    print("{:2}. Best guess: {:10.5f}".format(i, eigenvalue.item()))
    i += 1

print("Final error {:.5f}".format((eigenvalue - max_eigenvalue).abs()))



import torch
from torch import nn

from .exceptions import InManifoldError
from .utils import _extra_repr


def project(x):
    return x / x.norm(dim=-1, keepdim=True)


def uniform_init_sphere_(x, r=1.0):
    r"""Samples a point uniformly on the sphere into the tensor ``x``.
    If ``x`` has :math:`d > 1` dimensions, the first :math:`d-1` dimensions
    are treated as batch dimensions.
    """
    with torch.no_grad():
        x.normal_()
        x.data = r * project(x)
    return x


def _in_sphere(x, r, eps):
    norm = x.norm(dim=-1)
    rs = torch.full_like(norm, r)
    return (torch.norm(norm - rs, p=float("inf")) < eps).all()


class sinc_class(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        # Hardocoded for float, will do for now
        ret = torch.sin(x) / x
        ret[x.abs() < 1e-45] = 1.0
        return ret

    @staticmethod
    def backward(ctx, grad_output):
        (x,) = ctx.saved_tensors
        ret = torch.cos(x) / x - torch.sin(x) / (x * x)
        ret[x.abs() < 1e-10] = 0.0
        return ret * grad_output


sinc = sinc_class.apply


class SphereEmbedded(nn.Module):
    def __init__(self, size, radius=1.0):
        r"""
        Sphere as the orthogonal projection from
        :math:`\mathbb{R}^n` to :math:`\mathbb{S}^{n-1}`, that is,
        :math:`x \mapsto \frac{x}{\lVert x \rVert}`.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            radius (float): Optional.
                Radius of the sphere. A positive number. Default: ``1.``
        """
        super().__init__()
        self.n = size[-1]
        self.tensorial_size = size[:-1]
        self.radius = SphereEmbedded.parse_radius(radius)

    @staticmethod
    def parse_radius(radius):
        if radius <= 0.0:
            raise ValueError(
                "The radius has to be a positive real number. Got {}".format(radius)
            )
        return radius

    def forward(self, x):
        return self.radius * project(x)

    def right_inverse(self, x, check_in_manifold=True):
        if check_in_manifold and not self.in_manifold(x):
            raise InManifoldError(x, self)
        return x / self.radius

    def in_manifold(self, x, eps=1e-5):
        r"""
        Checks that a vector is on the sphere.

        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor): The vector to be checked.
            eps (float): Optional. Threshold at which the norm is considered
                    to be equal to `1`.
                    Default: ``1e-5``
        """
        return _in_sphere(x, self.radius, eps)

    def sample(self):
        r"""
        Returns a uniformly sampled vector on the sphere.
        """
        x = torch.empty(*(self.tensorial_size) + (self.n,))
        return uniform_init_sphere_(x, r=self.radius)

    def extra_repr(self):
        return _extra_repr(
            n=self.n, radius=self.radius, tensorial_size=self.tensorial_size
        )


class Sphere(nn.Module):
    def __init__(self, size, radius=1.0):
        r"""
        Sphere as a map from the tangent space onto the sphere using the
        exponential map.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            radius (float): Optional.
                Radius of the sphere. A positive number. Default: ``1.``
        """
        super().__init__()
        self.n = size[-1]
        self.tensorial_size = size[:-1]
        self.radius = Sphere.parse_radius(radius)
        self.register_buffer("base", uniform_init_sphere_(torch.empty(*size)))

    @staticmethod
    def parse_radius(radius):
        if radius <= 0.0:
            raise ValueError(
                "The radius has to be a positive real number. Got {}".format(radius)
            )
        return radius

    def frame(self, x, v):
        projection = (v.unsqueeze(-2) @ x.unsqueeze(-1)).squeeze(-1)
        v = v - projection * x
        return v

    def forward(self, v):
        x = self.base
        # Project v onto {<x,v> = 0}
        v = self.frame(x, v)
        vnorm = v.norm(dim=-1, keepdim=True)
        return self.radius * (torch.cos(vnorm) * x + sinc(vnorm) * v)

    def right_inverse(self, x, check_in_manifold=True):
        if check_in_manifold and not self.in_manifold(x):
            raise InManifoldError(x, self)
        with torch.no_grad():
            x = x / self.radius
            self.base.copy_(x)
        return torch.zeros_like(x)

    def in_manifold(self, x, eps=1e-5):
        r"""
        Checks that a vector is on the sphere.

        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor): The vector to be checked.
            eps (float): Optional. Threshold at which the norm is considered
                    to be equal to `1`.
                    Default: ``1e-5``
        """
        return _in_sphere(x, self.radius, eps)

    def sample(self):
        r"""
        Returns a uniformly sampled vector on the sphere.
        """
        device = self.base.device
        dtype = self.base.dtype
        x = torch.empty(*(self.tensorial_size) + (self.n,), device=device, dtype=dtype)
        return uniform_init_sphere_(x, r=self.radius)

    def extra_repr(self):
        return _extra_repr(
            n=self.n, radius=self.radius, tensorial_size=self.tensorial_size
        )



from .pssdlowrank import PSSDLowRank
from .exceptions import VectorError, NonSquareError
from .utils import _extra_repr


class PSSD(PSSDLowRank):
    def __init__(self, size, triv="expm"):
        r"""
        Manifold of symmetric positive semidefinite matrices

        Args:
            size (torch.size): Size of the tensor to be parametrized
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                matrices surjectively. This is used to optimize the :math:`Q` in the eigenvalue
                decomposition. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size, PSSD.rank(size), triv)

    @classmethod
    def rank(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        n, k = size[-2:]
        if n != k:
            raise NonSquareError(cls.__name__, size)
        return n

    def extra_repr(self):
        return _extra_repr(n=self.n, tensorial_size=self.tensorial_size)



import torch

from .stiefel import Stiefel


class Grassmannian(Stiefel):
    def __init__(self, size, triv="expm"):
        r"""
        Grassmannian manifold as a projection from the orthogonal
        matrices :math:`\operatorname{St}(n,k)`.
        The metric considered is the canonical.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size=size, triv=triv)

    def frame(self, X):
        k = X.size(-1)
        size_z = X.size()[:-2] + (k, k)
        Z = X.new_zeros(*size_z)
        X = torch.cat([Z, X[..., k:, :]], dim=-2)
        return super().frame(X)



import torch

try:
    from torch.linalg import qr
except ImportError:
    from torch import qr


from .utils import transpose, _extra_repr
from .so import SO, _has_orthonormal_columns

from .exceptions import VectorError, InManifoldError


class Stiefel(SO):
    def __init__(self, size, triv="expm"):
        r"""
        Manifold of rectangular orthogonal matrices parametrized as a projection
        onto the first :math:`k` columns from the space of square orthogonal matrices
        :math:`\operatorname{SO}(n)`. The metric considered is the canonical.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size=Stiefel.size_so(size), triv=triv, lower=True)
        self.k = min(size[-1], size[-2])
        self.transposed = size[-2] < size[-1]

    @classmethod
    def size_so(cls, size):
        if len(size) < 2:
            raise VectorError(cls.__name__, size)
        size_so = list(size)
        size_so[-1] = size_so[-2] = max(size[-1], size[-2])
        return tuple(size_so)

    def frame(self, X):
        n, k = X.size(-2), X.size(-1)
        size_z = X.size()[:-2] + (n, n - k)
        return torch.cat([X, X.new_zeros(*size_z)], dim=-1)

    @transpose
    def forward(self, X):
        X = self.frame(X)
        X = super().forward(X)
        return X[..., : self.k]

    @transpose
    def right_inverse(self, X, check_in_manifold=True):
        if check_in_manifold and not self.in_manifold(X):
            raise InManifoldError(X, self)
        if self.n != self.k:
            # N will be a completion of X to an orthogonal basis of R^n
            N = X.new_empty(*(self.tensorial_size + (self.n, self.n - self.k)))
            with torch.no_grad():
                N.normal_()
                # We assume for now that X is orthogonal.
                # This will be checked in super().right_inverse()
                # Project N onto the orthogonal complement to X
                # We iterate this twice for this algorithm to be numerically stable
                # This is standard, as done in some stochastic SVD algorithms
                for _ in range(2):
                    N = N - X @ (X.transpose(-2, -1) @ N)
                    # And make it an orthonormal base of the image
                    N = qr(N).Q
                X = torch.cat([X, N], dim=-1)
        return super().right_inverse(X, check_in_manifold=False)[..., : self.k]

    def in_manifold(self, X, eps=1e-4):
        r"""
        Checks that a matrix is in the manifold.

        For tensors with more than 2 dimensions the first dimensions are
        treated as batch dimensions.

        Args:
            X (torch.Tensor): The matrix to be checked
            eps (float): Optional. Tolerance to numerical errors.
                Default: ``1e-4``
        """
        if X.size(-1) > X.size(-2):
            X = X.transpose(-2, -1)
        if X.size() != self.tensorial_size + (self.n, self.k):
            return False
        return _has_orthonormal_columns(X, eps)

    def sample(self, distribution="uniform", init_=None):
        r"""
        Returns a randomly sampled orthogonal matrix according to the specified
        ``distribution``. The options are:

            - ``"uniform"``: Samples a tensor distributed according to the Haar measure
              on :math:`\operatorname{SO}(n)`

            - ``"torus"``: Samples a block-diagonal skew-symmetric matrix.
              The blocks are of the form
              :math:`\begin{pmatrix} 0 & b \\ -b & 0\end{pmatrix}` where :math:`b` is
              distributed according to ``init_``. This matrix will be then projected
              onto :math:`\operatorname{SO}(n)` using ``self.triv``

        .. note

            The ``"torus"`` initialization is particularly useful in recurrent kernels
            of RNNs

        Args:
            distribution (string): Optional. One of ``["uniform", "torus"]``.
                    Default: ``"uniform"``
            init\_ (callable): Optional. To be used with the ``"torus"`` option.
                    A function that takes a tensor and fills it in place according
                    to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: :math:`\operatorname{Uniform}(-\pi, \pi)`
        """
        X = super().sample(distribution, init_)
        if not self.transposed:
            return X[..., : self.k]
        else:
            return X[..., : self.k, :]

    def extra_repr(self):
        return _extra_repr(
            n=self.n,
            k=self.k,
            tensorial_size=self.tensorial_size,
            triv=self.triv,
            transposed=self.transposed,
        )



# Tests for the Sphere
from unittest import TestCase

from geotorch.sphere import Sphere, SphereEmbedded


class TestSphere(TestCase):
    def test_construction(self):
        # Negative curvature
        with self.assertRaises(ValueError):
            Sphere(size=(5,), radius=-1.0)
        with self.assertRaises(ValueError):
            SphereEmbedded(size=(4,), radius=-1.0)

    def test_repr(self):
        print(SphereEmbedded(size=(3,)))
        print(Sphere(size=(3,)))



import torch

from .symmetric import SymF
from .fixedrank import softplus_epsilon, inv_softplus_epsilon


class PSSDFixedRank(SymF):
    fs = {"softplus": (softplus_epsilon, inv_softplus_epsilon)}

    def __init__(self, size, rank, f="softplus", triv="expm"):
        r"""
        Manifold of symmetric positive semidefinite matrices of rank :math:`r`.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            rank (int): Rank of the matrices.
                It has to be less or equal to
                :math:`\min(\texttt{size}[-1], \texttt{size}[-2])`
            f (str or callable or pair of callables): Optional. Either:

                - ``"softplus"``

                - A callable that maps real numbers to the interval :math:`(0, \infty)`

                - A pair of callables such that the first maps the real numbers to
                  :math:`(0, \infty)` and the second is a (right) inverse of the first

                Default: ``"softplus"``
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`Q` in the eigenvalue
                decomposition. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size, rank, PSSDFixedRank.parse_f(f), triv)

    @staticmethod
    def parse_f(f):
        if f in PSSDFixedRank.fs.keys():
            return PSSDFixedRank.fs[f]
        elif callable(f):
            return f, None
        elif isinstance(f, tuple) and callable(f[0]) and callable(f[1]):
            return f
        else:
            raise ValueError(
                "Argument f was not recognized and is "
                "not callable or a pair of callables. "
                "Should be one of {}. Found {}".format(list(PSSDFixedRank.fs.keys()), f)
            )

    def in_manifold_eigen(self, L, eps=1e-6):
        r"""
        Checks that an ordered vector of eigenvalues values is in the manifold.

        For tensors with more than 1 dimension the first dimensions are
        treated as batch dimensions.

        Args:
            L (torch.Tensor): Vector of eigenvalues
            eps (float): Optional. Threshold at which the eigenvalues are
                considered to be zero
                Default: ``1e-6``
        """
        return (
            super().in_manifold_eigen(L, eps)
            and (L[..., : self.rank] >= eps).all().item()
        )

    def sample(self, init_=torch.nn.init.xavier_normal_, factorized=True, eps=5e-6):
        r"""
        Returns a randomly sampled matrix on the manifold as

        .. math::

            WW^\intercal \qquad W_{i,j} \sim \texttt{init_}

        If the sampled matrix has more than `self.rank` small singular values, the
        smallest ones are clamped to be at least ``eps`` in absolute value.


        The output of this method can be used to initialize a parametrized tensor as::

            >>> layer = nn.Linear(20, 20)
            >>> M = PSSD(layer.weight.size())
            >>> geotorch.register_parametrization(layer, "weight", M)
            >>> layer.weight = M.sample()

        Args:
            init\_ (callable): Optional.
                    A function that takes a tensor and fills it in place according
                    to some distribution. See
                    `torch.init <https://pytorch.org/docs/stable/nn.init.html>`_.
                    Default: ``torch.nn.init.xavier_normal_``
            factorized (bool): Optional. Return the tuple :math:`(\Lambda, Q)` with an
                    eigenvalue decomposition of the sampled matrix. This can also be used
                    to initialize the layer.
                    Default: ``True``
            eps (float): Optional. Minimum eigenvalue of the sampled matrix.
                    Default: ``5e-6``
        """
        L, Q = super().sample(factorized=True, init_=init_)
        with torch.no_grad():
            # S >= 0, as given by torch.linalg.eigvalsh()
            small = L < eps
            L[small] = eps
        if factorized:
            return L, Q
        else:
            # Project onto the manifold
            Qt = Q.transpose(-2, -1)
            # Multiply the three of them as Q\LambdaQ^T
            return Q @ (L.unsqueeze(-1).expand_as(Qt) * Qt)



class VectorError(ValueError):
    def __init__(self, name, size):
        super().__init__(
            "Cannot instantiate {} on a tensor of less than 2 dimensions. "
            "Got a tensor of size {}".format(name, size)
        )


class InverseError(ValueError):
    def __init__(self, M):
        super().__init__(
            "Cannot initialize the parametrization {} as no inverse for the function "
            "{} was specified in the constructor".format(M, M.f.__name__)
        )


class NonSquareError(ValueError):
    def __init__(self, name, size):
        super().__init__(
            "The {} parametrization can just be applied to square matrices. "
            "Got a tensor of size {}".format(name, size)
        )


class RankError(ValueError):
    def __init__(self, n, k, rank):
        super().__init__(
            "The rank has to be 1 <= rank <= min({}, {}). Found {}".format(n, k, rank)
        )


class InManifoldError(ValueError):
    def __init__(self, X, M):
        super().__init__("Tensor not contained in {}. Got\n{}".format(M, X))



from setuptools import setup, find_packages
import os
import re

TEST_REQUIRES = ["pytest"]

DEV_REQUIRES = TEST_REQUIRES + [
    "black",
    "flake8",
    "sphinx",
    "sphinx-autodoc-typehints",
    "sphinx-rtd-theme",
    "sphinxcontrib-spelling",
    "codecov",
]

classifiers = [
    "Development Status :: 3 - Alpha",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "License :: OSI Approved :: MIT License",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
]

# Get the long description from the README file
with open("README.rst", "r", encoding="utf8") as fh:
    long_description = fh.read()

# Get version string from module
init_path = os.path.join(os.path.dirname(__file__), "geotorch/__init__.py")
with open(init_path, "r", encoding="utf8") as f:
    version = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M).group(1)

setup(
    name="geotorch",
    version=version,
    description="Constrained Optimization and Manifold Optimization in Pytorch",
    author="Mario Lezcano Casado",
    author_email="lezcano-93@hotmail.com",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/Lezcano/geotorch",
    classifiers=classifiers,
    keywords=["Constrained Optimization", "Optimization on Manifolds", "Pytorch"],
    packages=find_packages(),
    python_requires=">=3.5",
    install_requires=["torch>=1.5"],
    extras_require={"dev": DEV_REQUIRES, "test": TEST_REQUIRES},
)



from unittest import TestCase

import torch
from geotorch.product import ProductManifold
from geotorch.so import SO


class TestManifold(TestCase):
    def test_product_manifold(self):
        # Should not throw
        SO3SO3 = ProductManifold([SO((3, 3)), SO((3, 3))])

        # A tuple should work as well
        SO3SO3 = ProductManifold((SO((3, 3)), SO((3, 3))))

        # Forward should work
        X = (torch.rand(3, 3), torch.rand(3, 3))
        Y1, Y2 = SO3SO3(X)



import math
from unittest import TestCase
import torch

from geotorch.linalg.expm import expm, taylor_approx


class TestLinalg(TestCase):
    def taylor(self, X, deg):
        assert X.size(-1) == X.size(-2)
        n = X.size(-1)
        Id = torch.eye(n, n, dtype=X.dtype, device=X.device)
        if X.ndimension() > 2:
            Id = Id.expand_as(X)
        acc = Id
        last = Id
        for i in range(1, deg + 1):
            last = last @ X / float(i)
            acc = acc + last
        return acc

    def scale_square(self, X):
        """
        Scale-squaring trick
        """
        norm = X.norm()
        if norm < 0.5:
            return self.taylor(X, 12)

        k = int(math.ceil(math.log2(float(norm)))) + 2
        X = X * (2 ** -k)
        E = self.taylor(X, 18)
        for _ in range(k):
            E = torch.mm(E, E)
        return E

    def assertIsCloseSquare(self, X, Y, places=4):
        self.assertEqual(X.ndimension(), 2)
        self.assertEqual(X.size(0), X.size(1))
        self.assertAlmostEqual(torch.norm(X - Y).item(), 0.0, places=places)

    def compare_f(self, f_batching, f_simple, allows_batches, dtype, gradients=False):
        # Test expm without batching
        for _ in range(8):
            A = torch.rand(10, 10, dtype=dtype)
            if gradients:
                G = torch.rand(10, 10, dtype=dtype)
                B1 = f_batching(A, G)
                B2 = f_simple(A, G)
                self.assertIsCloseSquare(B1, B2, places=2)
            else:
                B1 = f_batching(A)
                B2 = f_simple(A)
                self.assertIsCloseSquare(B1, B2, places=3)

        # Test batching
        for _ in range(3):
            len_shape = torch.randint(1, 4, (1,))
            shape_batch = torch.randint(1, 5, size=(len_shape,))
            shape = list(shape_batch) + [8, 8]
            A = torch.rand(*shape, dtype=dtype)
            if gradients:
                G = torch.rand(*shape, dtype=dtype)
                B1 = f_batching(A, G)
            else:
                B1 = f_batching(A)
            if allows_batches:
                B2 = f_simple(A)
                self.assertEqual(B1.size(), A.size())
                self.assertEqual(B2.size(), A.size())

                # sample a few coordinates and evaluate the equality
                # of those elements in the batch
                for _ in range(3):
                    coords = [
                        torch.randint(low=0, high=s, size=(1,)).item()
                        for s in shape_batch
                    ]
                    coords = coords + [...]
                    self.assertIsCloseSquare(B1[coords], B2[coords], places=3)
            else:
                # sample a few coordinates and evaluate the equality
                # of those elements in the batch
                for _ in range(3):
                    coords = [
                        torch.randint(low=0, high=s, size=(1,)).item()
                        for s in shape_batch
                    ]
                    coords = coords + [...]
                    if gradients:
                        self.assertIsCloseSquare(
                            B1[coords], f_simple(A[coords], G[coords]), places=2
                        )
                    else:
                        self.assertIsCloseSquare(
                            B1[coords], f_simple(A[coords]), places=3
                        )

    def test_expm(self):
        with torch.random.fork_rng(devices=range(torch.cuda.device_count())):
            torch.random.manual_seed(8888)
            # Test different Taylor approximations
            degs = [1, 2, 4, 8, 12, 18]
            for deg in degs:
                for dtype in [torch.float, torch.double]:
                    self.compare_f(
                        lambda X: taylor_approx(X, deg),
                        lambda X: self.taylor(X, deg),
                        allows_batches=True,
                        dtype=dtype,
                    )

            # Test the main function
            for dtype in [torch.float, torch.double]:
                self.compare_f(
                    expm, self.scale_square, allows_batches=False, dtype=dtype
                )

            # Test the gradients
            def diff(f):
                def wrap(A, G):
                    A.requires_grad_()
                    return torch.autograd.grad([f(A)], [A], [G])[0]

                return wrap

            for dtype in [torch.float, torch.double]:
                self.compare_f(
                    diff(expm),
                    diff(self.scale_square),
                    allows_batches=False,
                    dtype=dtype,
                    gradients=True,
                )

    def test_errors(self):
        with self.assertRaises(ValueError):
            expm(torch.empty(3, 4))
        with self.assertRaises(ValueError):
            expm(torch.empty(1, 4))



import torch
from .symmetric import SymF


class PSSDLowRank(SymF):
    def __init__(self, size, rank, triv="expm"):
        r"""
        Variety of the symmetric positive semidefinite matrices of rank
        at most :math:`r`.

        Args:
            size (torch.size): Size of the tensor to be parametrized
            rank (int): Rank of the matrices.
                It has to be less or equal to
                :math:`\min(\texttt{size}[-1], \texttt{size}[-2])`
            triv (str or callable): Optional.
                A map that maps skew-symmetric matrices onto the orthogonal matrices
                surjectively. This is used to optimize the :math:`Q` in the eigenvalue
                decomposition. It can be one of ``["expm", "cayley"]`` or a custom
                callable. Default: ``"expm"``
        """
        super().__init__(size, rank, f=(torch.abs, torch.abs), triv=triv)



# Integration tests for all the manifold
from unittest import TestCase
import itertools
import types

import torch
import torch.nn as nn
import geotorch.parametrize as P

import geotorch
from geotorch.skew import Skew
from geotorch.symmetric import Symmetric
from geotorch.so import SO
from geotorch.stiefel import Stiefel
from geotorch.grassmannian import Grassmannian
from geotorch.lowrank import LowRank
from geotorch.fixedrank import FixedRank
from geotorch.psd import PSD
from geotorch.pssd import PSSD
from geotorch.pssdlowrank import PSSDLowRank
from geotorch.pssdfixedrank import PSSDFixedRank
from geotorch.glp import GLp
from geotorch.almostorthogonal import AlmostOrthogonal
from geotorch.sphere import Sphere, SphereEmbedded

from geotorch.utils import update_base


def dicts_product(**kwargs):
    """ Returns a product of all the lists of the keys """
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


class TestIntegration(TestCase):
    def sizes(self, square):
        sizes = []
        if not torch.cuda.is_available():
            sizes = [(i, i) for i in range(1, 11)]
            if not square:
                sizes.extend(
                    [
                        (i, j)
                        for i, j in itertools.product(range(1, 5), range(1, 5))
                        if i != j
                    ]
                )
                sizes.extend(
                    [(1, 7), (2, 7), (1, 8), (2, 8), (7, 1), (7, 2), (8, 1), (8, 2)]
                )
        else:
            sizes.extend([(256, 256), (512, 512)])
            if not square:
                sizes.extend([(256, 128), (128, 512), (1024, 512)])
        return sizes

    def ranks(self):
        return [1, 3, 7]

    def lambdas(self):
        return [0.01, 0.5, 1.0]

    def radii(self):
        return [0.01, 1.0, 2.0, 10.0]

    def devices(self):
        if torch.cuda.is_available():
            return [torch.device("cuda")]
        else:
            return [torch.device("cpu")]

    def test_vector_spaces(self):
        self._test_manifolds(
            [Skew, Symmetric, geotorch.skew, geotorch.symmetric],
            [],
            dicts_product(lower=[True, False]),
            self.devices(),
            self.sizes(square=True),
            initialize=False,
        )

    def test_so(self):
        self._test_manifolds(
            [SO],
            dicts_product(distribution=["uniform", "torus"]),
            [{}],
            self.devices(),
            self.sizes(square=True),
        )

    def test_orthogonal(self):
        self._test_manifolds(
            [Stiefel, Grassmannian, geotorch.orthogonal, geotorch.grassmannian],
            dicts_product(distribution=["uniform", "torus"]),
            [{}],
            self.devices(),
            self.sizes(square=False),
        )

    def test_rank(self):
        self._test_manifolds(
            [LowRank, FixedRank, geotorch.low_rank, geotorch.fixed_rank],
            dicts_product(factorized=[True, False]),
            dicts_product(rank=self.ranks()),
            self.devices(),
            self.sizes(square=False),
        )

    def test_psd_and_glp(self):
        self._test_manifolds(
            [
                PSD,
                PSSD,
                GLp,
                geotorch.positive_definite,
                geotorch.positive_semidefinite,
                geotorch.invertible,
            ],
            dicts_product(factorized=[True, False]),
            [{}],
            self.devices(),
            self.sizes(square=True),
        )

    def test_pssd_rank(self):
        self._test_manifolds(
            [
                PSSDLowRank,
                PSSDFixedRank,
                geotorch.positive_semidefinite_low_rank,
                geotorch.positive_semidefinite_fixed_rank,
            ],
            dicts_product(factorized=[True, False]),
            dicts_product(rank=self.ranks()),
            self.devices(),
            self.sizes(square=True),
        )

    def test_almost_orthogonal(self):
        self._test_manifolds(
            [AlmostOrthogonal, geotorch.almost_orthogonal],
            dicts_product(factorized=[True, False], distribution=["uniform", "torus"]),
            dicts_product(lam=self.lambdas(), f=list(AlmostOrthogonal.fs.keys())),
            self.devices(),
            self.sizes(square=True),
        )

    def test_sphere(self):
        self._test_manifolds(
            [Sphere, SphereEmbedded, geotorch.sphere],
            [{}],
            dicts_product(radius=self.radii()),
            self.devices(),
            self.sizes(square=False),
        )

    def _test_manifolds(
        self, Ms, argss_sample, argss_constr, devices, sizes, initialize=False
    ):
        with torch.random.fork_rng(devices=range(torch.cuda.device_count())):
            torch.random.manual_seed(8888)
            for M, args_sample, args_constr, device, size in itertools.product(
                Ms, argss_sample, argss_constr, devices, sizes
            ):
                if "rank" in args_constr and args_constr["rank"] > min(size):
                    continue
                self._test_manifold(
                    M, args_sample, args_constr, device, size, initialize
                )

    def _test_manifold(self, M, args_sample, args_constr, device, size, initialize):
        # Test Linear
        layer = nn.Linear(*size)
        input_ = torch.rand(3, size[0]).to(device)
        old_size = layer.weight.size()
        # Somewhat dirty but will do
        if isinstance(M, types.FunctionType):
            M(layer, "weight", **args_constr)
        else:
            P.register_parametrization(
                layer, "weight", M(size=layer.weight.size(), **args_constr)
            )
        layer = layer.to(device)
        # Check that it does not change the size of the layer
        self.assertEqual(old_size, layer.weight.size(), msg=f"{layer}")
        self._test_training(layer, args_sample, input_, initialize)

        # Just for the smaller ones, for the large ones this is just too expensive
        if min(size) < 100:
            # Test Convolutionar (tensorial)
            layer = nn.Conv2d(5, 4, size)
            input_ = torch.rand(6, 5, size[0] + 7, size[1] + 3).to(device)
            old_size = layer.weight.size()
            # Somewhat dirty but will do
            if isinstance(M, types.FunctionType):
                M(layer, "weight", **args_constr)
            else:
                P.register_parametrization(
                    layer, "weight", M(size=layer.weight.size(), **args_constr)
                )
            layer = layer.to(device)
            # Check that it does not change the size of the layer
            self.assertEqual(old_size, layer.weight.size(), msg=f"{layer}")
            self._test_training(layer, args_sample, input_, initialize)

    def matrix_from_factor_svd(self, U, S, V):
        Vt = V.transpose(-2, -1)
        # Multiply the three of them, S as a diagonal matrix
        return U @ (S.unsqueeze(-1).expand_as(Vt) * Vt)

    def matrix_from_factor_eigen(self, L, Q):
        Qt = Q.transpose(-2, -1)
        # Multiply the three of them as Q\LambdaQ^T
        return Q @ (L.unsqueeze(-1).expand_as(Qt) * Qt)

    def matrix_from_factor(self, X, M):
        transpose = hasattr(M, "transposed") and M.transposed
        if not isinstance(X, tuple):
            return X
        elif len(X) == 2:
            X = self.matrix_from_factor_eigen(X[0], X[1])
        else:
            X = self.matrix_from_factor_svd(X[0], X[1], X[2])
        if transpose:
            return X.transpose(-2, -1)
        else:
            return X

    def _test_training(self, layer, args_sample, input_, initialize):
        msg = f"{layer}\n{args_sample}"
        M = layer.parametrizations.weight[0]
        if initialize:
            initial_size = layer.weight.size()
            X = M.sample(**args_sample)
            self.assertTrue(M.in_manifold(X), msg=msg)
            layer.weight = X
            with P.cached():
                # Compute the product if it is factorized
                X_matrix = self.matrix_from_factor(X, M).to(layer.weight.device)
                # The sampled matrix should not have a gradient
                self.assertFalse(X_matrix.requires_grad)
                # Size does not change
                self.assertEqual(initial_size, layer.weight.size(), msg=msg)
                # Tha initialisation initialisation is equal to what we passed
                self.assertTrue(
                    torch.allclose(layer.weight, X_matrix, atol=1e-5), msg=msg
                )

        # Take a couple SGD steps
        optim = torch.optim.SGD(layer.parameters(), lr=1e-3)
        for i in range(3):
            with P.cached():
                loss = layer(input_).mean()
            optim.zero_grad()
            loss.backward()
            optim.step()
            # The layer stays in the manifold while being optimised
            self.assertTrue(M.in_manifold(layer.weight), msg=f"i:{i}\n" + msg)

        with P.cached():
            weight_old = layer.weight
            update_base(layer, "weight")
            # After changing the base, the weight stays the same
            self.assertTrue(
                torch.allclose(layer.weight, weight_old, atol=1e-6), msg=msg
            )



"""
Slightly more advanced example of usage of GeoTorch

Implements a constrained RNN to classify MNIST processing the images one pixel at a time
A good result for this task for size 170 would be 98.0% accuracy with orthogonal constraints
and 98.5% for the almostorthogonal. Lowrank is here for demonstration purposes, it should
not perform as well as the other two

The GeoTorch code happens in `ExpRNNCell.__init__`, `ExpRNNCell.reset_parameters` and line 132.
The rest of the code is normal PyTorch.
Lines 167-176 show how to assign different learning rates to parametrized weights
"""

import torch
import torch.nn as nn
import math
import argparse
from torchvision import datasets, transforms

import geotorch
from geotorch.so import torus_init_

parser = argparse.ArgumentParser(description="Exponential Layer MNIST Task")
parser.add_argument("--batch_size", type=int, default=128)
parser.add_argument("--hidden_size", type=int, default=170)
parser.add_argument("--epochs", type=int, default=70)
parser.add_argument("--lr", type=float, default=1e-3)
parser.add_argument("--lr_orth", type=float, default=1e-4)
parser.add_argument("--permute", action="store_true")
parser.add_argument(
    "--constraints",
    choices=["orthogonal", "lowrank", "almostorthogonal"],
    default="orthogonal",
    type=str,
)
parser.add_argument(
    "--f",
    choices=["scaled_sigmoid", "tanh", "sin"],
    default="scaled_sigmoid",
    type=str,
)
parser.add_argument("--r", type=float, default=0.1)


args = parser.parse_args()

n_classes = 10
batch_size = args.batch_size
hidden_size = args.hidden_size
epochs = args.epochs
device = torch.device("cuda")


class modrelu(nn.Module):
    def __init__(self, features):
        super(modrelu, self).__init__()
        self.features = features
        self.b = nn.Parameter(torch.Tensor(self.features))
        self.reset_parameters()

    def reset_parameters(self):
        self.b.data.uniform_(-0.01, 0.01)

    def forward(self, inputs):
        norm = torch.abs(inputs)
        biased_norm = norm + self.b
        magnitude = nn.functional.relu(biased_norm)
        phase = torch.sign(inputs)

        return phase * magnitude


class ExpRNNCell(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(ExpRNNCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.recurrent_kernel = nn.Linear(hidden_size, hidden_size, bias=False)
        self.input_kernel = nn.Linear(input_size, hidden_size)
        self.nonlinearity = modrelu(hidden_size)

        # Make recurrent_kernel orthogonal
        if args.constraints == "orthogonal":
            geotorch.orthogonal(self.recurrent_kernel, "weight")
        elif args.constraints == "lowrank":
            geotorch.low_rank(self.recurrent_kernel, "weight", hidden_size)
        elif args.constraints == "almostorthogonal":
            geotorch.almost_orthogonal(self.recurrent_kernel, "weight", args.r, args.f)
        else:
            raise ValueError("Unexpected constraints. Got {}".format(args.constraints))

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_normal_(self.input_kernel.weight.data, nonlinearity="relu")

        # Initialize the recurrent kernel à la Cayley, as having a block-diagonal matrix
        # seems to help in classification problems

        def init_(x):
            x.uniform_(0.0, math.pi / 2.0)
            c = torch.cos(x.data)
            x.data = -torch.sqrt((1.0 - c) / (1.0 + c))

        K = self.recurrent_kernel
        # We initialize it by assigning directly to it from a sampler
        K.weight = torus_init_(K.weight, init_=init_)

    def default_hidden(self, input_):
        return input_.new_zeros(input_.size(0), self.hidden_size, requires_grad=False)

    def forward(self, input_, hidden):
        input_ = self.input_kernel(input_)
        hidden = self.recurrent_kernel(hidden)
        out = input_ + hidden
        return self.nonlinearity(out)


class Model(nn.Module):
    def __init__(self, hidden_size, permute):
        super(Model, self).__init__()
        self.permute = permute
        if self.permute:
            self.register_buffer("permutation", torch.randperm(784))
        self.rnn = ExpRNNCell(1, hidden_size)
        self.lin = nn.Linear(hidden_size, n_classes)
        self.loss_func = nn.CrossEntropyLoss()

    def forward(self, inputs):
        if self.permute:
            inputs = inputs[:, self.permutation]
        out_rnn = self.rnn.default_hidden(inputs[:, 0, ...])
        with geotorch.parametrize.cached():
            for input in torch.unbind(inputs, dim=1):
                out_rnn = self.rnn(input.unsqueeze(dim=1), out_rnn)
        return self.lin(out_rnn)

    def loss(self, logits, y):
        return self.loss_func(logits, y)

    def correct(self, logits, y):
        return torch.eq(torch.argmax(logits, dim=1), y).float().sum()


def main():
    # Load data
    kwargs = {
        "batch_size": batch_size,
        "num_workers": 1,
        "pin_memory": True,
        "shuffle": True,
    }
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST(
            "./mnist", train=True, download=True, transform=transforms.ToTensor()
        ),
        **kwargs
    )
    test_loader = torch.utils.data.DataLoader(
        datasets.MNIST("./mnist", train=False, transform=transforms.ToTensor()),
        **kwargs
    )

    # Model and optimizers
    model = Model(hidden_size, args.permute).to(device)
    model.train()

    p_orth = model.rnn.recurrent_kernel
    orth_params = p_orth.parameters()
    non_orth_params = (
        param for param in model.parameters() if param not in set(p_orth.parameters())
    )

    optim = torch.optim.RMSprop(
        [{"params": non_orth_params}, {"params": orth_params, "lr": args.lr_orth}],
        lr=args.lr,
    )

    best_test_acc = 0.0
    for epoch in range(epochs):
        processed = 0
        for batch_idx, (batch_x, batch_y) in enumerate(train_loader):
            batch_x, batch_y = batch_x.to(device).view(-1, 784), batch_y.to(device)

            optim.zero_grad()
            logits = model(batch_x)
            loss = model.loss(logits, batch_y)
            loss.backward()
            optim.step()

            with torch.no_grad():
                correct = model.correct(logits, batch_y)

            processed += len(batch_x)
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.2f}%\tBest: {:.2f}%".format(
                    epoch,
                    processed,
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                    100 * correct / len(batch_x),
                    best_test_acc,
                )
            )

        model.eval()
        with torch.no_grad():
            test_loss = 0.0
            correct = 0.0
            for batch_x, batch_y in test_loader:
                batch_x, batch_y = batch_x.to(device).view(-1, 784), batch_y.to(device)
                logits = model(batch_x)
                test_loss += model.loss(logits, batch_y).float()
                correct += model.correct(logits, batch_y).float()

        test_loss /= len(test_loader)
        test_acc = 100 * correct / len(test_loader.dataset)
        best_test_acc = max(test_acc, best_test_acc)
        print(
            "\nTest set: Average loss: {:.4f}, Accuracy: {:.2f}%, Best Accuracy: {:.2f}%\n".format(
                test_loss, test_acc, best_test_acc
            )
        )

        model.train()


if __name__ == "__main__":
    main()



from unittest import TestCase

from geotorch.lowrank import LowRank
from geotorch.fixedrank import FixedRank


class TestLowRank(TestCase):
    def test_lowrank_errors(self):
        # rank always has to be <= min(n, k)
        for cls in [LowRank, FixedRank]:
            with self.assertRaises(ValueError):
                cls(size=(4, 3), rank=5)
            with self.assertRaises(ValueError):
                cls(size=(2, 3), rank=3)
            # Try to instantiate it in a vector rather than a matrix
            with self.assertRaises(ValueError):
                cls(size=(5,), rank=1)

        # On a non-callable
        with self.assertRaises(ValueError):
            FixedRank(size=(5, 3), rank=2, f=3)
        # On the wrong string
        with self.assertRaises(ValueError):
            FixedRank(size=(5, 3), rank=2, f="wrong")



from torch import nn
from .utils import _extra_repr
from .exceptions import InManifoldError


class Rn(nn.Module):
    def __init__(self, size):
        r"""
        Vector space of unconstrained vectors.

        Args:
            size (torch.size): Size of the tensor to be parametrized
        """
        super().__init__()
        self.n = size[-1]
        self.tensorial_size = size[:-1]

    def forward(self, X):
        return X

    def right_inverse(self, X, check_in_manifold=True):
        if check_in_manifold and not self.in_manifold(X):
            raise InManifoldError(X, self)
        return X

    def in_manifold(self, X):
        return X.size() == self.tensorial_size + (self.n,)

    def extra_repr(self):
        return _extra_repr(n=self.n, tensorial_size=self.tensorial_size)



from unittest import TestCase

from geotorch.almostorthogonal import AlmostOrthogonal


class TestLowRank(TestCase):
    def test_almostorthogonal_errors(self):
        with self.assertRaises(ValueError):
            AlmostOrthogonal(size=(5,), lam=1.0)
        # Not a predefined f
        with self.assertRaises(ValueError):
            AlmostOrthogonal(size=(5, 4), lam=1, f="fail")
        # Not callable
        with self.assertRaises(ValueError):
            AlmostOrthogonal(size=(5, 4), lam=1, f=3.0)
        # But a callable should work
        AlmostOrthogonal(size=(5, 4), lam=0.5)
        # Too large a lambda
        with self.assertRaises(ValueError):
            AlmostOrthogonal(size=(5, 4), lam=2)
        # Or too small
        with self.assertRaises(ValueError):
            AlmostOrthogonal(size=(5, 4), lam=-1.0)
